import os
import numpy as np

from os.path import join as opj

from ...core.basejob import SingleJob
from ...core.errors import FileError, JobError, ResultsError, PTError
from ...core.functions import config, log, parse_heredoc
from ...core.private import sha256, UpdateSysPath
from ...core.results import Results
from ...core.settings import Settings
from ...mol.molecule import Molecule
from ...mol.atom import Atom
from ...tools.kftools import KFFile
from ...tools.units import Units
from ...trajectories.rkffile import RKFTrajectoryFile
from ...tools.converters import vasp_output_to_ams
import subprocess


__all__ = ['AMSJob', 'AMSResults']


class AMSResults(Results):
    """A specialized |Results| subclass for accessing the results of |AMSJob|."""

    def __init__(self, *args, **kwargs):
        Results.__init__(self, *args, **kwargs)
        self.rkfs = {}


    def collect(self):
        """Collect files present in the job folder. Use parent method from |Results|, then create an instance of |KFFile| for each ``.rkf`` file present in the job folder. Collect these files in ``rkfs`` dictionary, with keys being filenames without ``.rkf`` extension.

        The information about ``.rkf`` files generated by engines is taken from the main ``ams.rkf`` file.

        This method is called automatically during the final part of the job execution and there is no need to call it manually.
        """
        Results.collect(self)

        rkfname = 'ams.rkf'
        if rkfname in self.files:
            main = KFFile(opj(self.job.path, rkfname))
            n = main[('EngineResults','nEntries')]
            for i in range(1, n+1):
                title =  main[('EngineResults','Title({})'.format(i))]
                files =  main[('EngineResults','Files({})'.format(i))].split('\x00')
                if files[0].endswith('.rkf'):
                    key = files[0][:-4]
                    self.rkfs[key] = KFFile(opj(self.job.path, files[0]))
            self.rkfs['ams'] = main

        else:
            log('WARNING: Main KF file {} not present in {}'.format(rkfname, self.job.path), 1)


    def refresh(self):
        """Refresh the contents of ``files`` list.

        Use the parent method from |Results|, then look at |KFFile| instances present in ``rkfs`` dictionary and check if they point to existing files. If not, try to reinstantiate them with current job path (that can happen while loading a pickled job after the entire job folder was moved).
        """
        Results.refresh(self)
        to_remove = []
        for key,val in self.rkfs.items():
            if not os.path.isfile(val.path):
                if os.path.dirname(val.path) != self.job.path:
                    guessnewpath = opj(self.job.path, os.path.basename(val.path))
                    if os.path.isfile(guessnewpath):
                        self.rkfs[key] = KFFile(guessnewpath)
                    else:
                        to_remove.append(key)
                else:
                    to_remove.append(key)
        for i in to_remove:
            del self.rkfs[i]


    def engine_names(self):
        """Return a list of all names of engine specific ``.rkf`` files. The identifier of the main result file (``'ams'``) is not present in the returned list, only engine specific names are listed.
        """
        self.refresh()
        ret = list(self.rkfs.keys())
        ret.remove('ams')
        return ret


    def rkfpath(self, file='ams'):
        """Return the absolute path of a chosen ``.rkf`` file.

        The *file* argument should be the identifier of the file to read. It defaults to ``'ams'``. To access a file called ``something.rkf`` you need to call this function with ``file='something'``. If there exists only one engine results ``.rkf`` file, you can call this function with ``file='engine'`` to access this file.
        """
        return self._access_rkf(lambda x: x.path, file)


    def readrkf(self, section, variable, file='ams'):
        """Read data from *section*/*variable* of a chosen ``.rkf`` file.

        The *file* argument should be the identifier of the file to read. It defaults to ``'ams'``. To access a file called ``something.rkf`` you need to call this function with ``file='something'``. If there exists only one engine results ``.rkf`` file, you can call this function with ``file='engine'`` to access this file.

        The type of the returned value depends on the type of *variable* defined inside KF file. It can be: single int, list of ints, single float, list of floats, single boolean, list of booleans or string.

        .. note::

            If arguments *section* or *variable* are incorrect (not present in the chosen file), the returned value is ``None``. Please mind the fact that KF files are case sensitive.

        """
        return self._access_rkf(lambda x: x.read(section, variable), file)


    def read_rkf_section(self, section, file='ams'):
        """Return a dictionary with all variables from a given *section* of a chosen ``.rkf`` file.

        The *file* argument should be the identifier of the file to read. It defaults to ``'ams'``. To access a file called ``something.rkf`` you need to call this function with ``file='something'``. If there exists only one engine results ``.rkf`` file, you can call this function with ``file='engine'`` to access this file.

        .. note::

            If *section* is not present in the chosen file, the returned value is an empty dictionary. Please mind the fact that KF files are case sensitive.

        """
        return self._access_rkf(lambda x: x.read_section(section), file)


    def get_rkf_skeleton(self, file='ams'):
        """Return a dictionary with the structure of a chosen ``.rkf`` file. Each key corresponds to a section name with the value being a set of variable names present in that section.

        The *file* argument should be the identifier of the file to read. It defaults to ``'ams'``. To access a file called ``something.rkf`` you need to call this function with ``file='something'``. If there exists only one engine results ``.rkf`` file, you can call this function with ``file='engine'`` to access this file.
        """
        return self._access_rkf(lambda x: x.get_skeleton(), file)


    def get_molecule(self, section, file='ams'):
        """Return a |Molecule| instance stored in a given *section* of a chosen ``.rkf`` file.

        The *file* argument should be the identifier of the file to read. It defaults to ``'ams'``. To access a file called ``something.rkf`` you need to call this function with ``file='something'``. If there exists only one engine results ``.rkf`` file, you can call this function with ``file='engine'`` to access this file.

        All data used by this method is taken from the chosen ``.rkf`` file. The ``molecule`` attribute of the corresponding job is ignored.
        """
        sectiondict = self.read_rkf_section(section, file)
        if sectiondict:
            return Molecule._mol_from_rkf_section(sectiondict)


    def get_input_molecule(self):
        """Return a |Molecule| instance with the initial coordinates.

        All data used by this method is taken from ``ams.rkf`` file. The ``molecule`` attribute of the corresponding job is ignored.
        """
        return self.get_molecule('InputMolecule', 'ams')


    def get_main_molecule(self):
        """Return a |Molecule| instance with the final coordinates.

        All data used by this method is taken from ``ams.rkf`` file. The ``molecule`` attribute of the corresponding job is ignored.
        """
        return self.get_molecule('Molecule', 'ams')


    def get_history_molecule(self, step):
        """Return a |Molecule| instance with coordinates taken from a particular *step* in the ``History`` section of ``ams.rkf`` file.

        All data used by this method is taken from ``ams.rkf`` file. The ``molecule`` attribute of the corresponding job is ignored.
        """
        if 'ams' in self.rkfs:
            main = self.rkfs['ams']
            if 'History' not in main:
                raise KeyError("'History' section not present in {}".format(main.path))
            n = main.read('History', 'nEntries')
            if step > n:
                raise KeyError("Step {} not present in 'History' section of {}".format(step, main.path))
            coords = main.read('History', f'Coords({step})')
            coords = [coords[i:i+3] for i in range(0,len(coords),3)]
            if ('History', f'SystemVersion({step})') in main:
                version = main.read('History', f'SystemVersion({step})')
                if 'SystemVersionHistory' in main:
                    if ('SystemVersionHistory', 'blockSize') in main:
                        blockSize = main.read('SystemVersionHistory', 'blockSize')
                    else:
                        blockSize = 1
                    block = (version - 1) // blockSize + 1
                    offset = (version - 1) % blockSize
                    system = main.read('SystemVersionHistory', f'SectionNum({block})', return_as_list=True)[offset]
                else:
                    system = version
                mol = self.get_molecule(f'ChemicalSystem({system})')
                molsrc = f'ChemicalSystem({system})'
            else:
                mol = self.get_main_molecule()
                molsrc = 'Molecule'
            if len(mol) != len(coords):
                raise ResultsError(f'Coordinates taken from "History%Coords({step})" have incompatible length with molecule from {molsrc} section')
            for at, c in zip(mol, coords):
                at.move_to(c, unit='bohr')

            if ('History', f'LatticeVectors('+str(step)+')') in main:
                lattice = Units.convert(main.read('History', f'LatticeVectors('+str(step)+')'), 'bohr', 'angstrom')
                mol.lattice = [tuple(lattice[j:j+3]) for j in range(0,len(lattice),3)]

            if all(('History', i) in main for i in [f'Bonds.Index({step})', f'Bonds.Atoms({step})', f'Bonds.Orders({step})']):
                mol.bonds = []
                index = main.read('History', f'Bonds.Index({step})')
                if not isinstance(index, list): index = [index]
                atoms = main.read('History', f'Bonds.Atoms({step})')
                if not isinstance(atoms, list): atoms = [atoms]
                orders = main.read('History', f'Bonds.Orders({step})')
                if not isinstance(orders, list): orders = [orders]
                for i in range(len(index)-1):
                    for j in range(index[i], index[i+1]):
                        mol.add_bond(mol[i+1], mol[atoms[j-1]], orders[j-1])
            return mol

    def get_history_variables(self, history_section='History') :
        """ Return a set of keynames stored in the specified history section of the ``ams.rkf`` file.

        The *history_section argument should be a string representing the name of the history section (``History`` or ``MDHistory``)*"""
        if not 'ams' in self.rkfs: return
        main = self.rkfs['ams']
        keylist = [var for sec,var in main if sec==history_section]
        # Now throw out all the last parts
        return set([key.split('(')[0] for key in keylist if len(key.split('('))>1])

    def get_history_property(self, varname, history_section='History') :
        """ Return the values of *varname* in the history section *history_section*."""
        if not 'ams' in self.rkfs: return
        main = self.rkfs['ams']
        if (history_section,'nScanCoord') in main: # PESScan
            nentries = main.read(history_section,'nScanCoord') 
            as_block = False
        elif (history_section,'nEntries') in main: 
            nentries = main.read(history_section,'nEntries')
            as_block = self._values_stored_as_blocks(main, varname, history_section)
        if as_block :
            nblocks = main.read(history_section,'nBlocks')
            values = [main.read(history_section,f"{varname}({iblock})",return_as_list=True) for iblock in range(1,nblocks+1)]
            values = [val for blockvals in values for val in blockvals if isinstance(blockvals,list)]
        else :
            values = [main.read(history_section,f"{varname}({step})") for step in range(1,nentries+1)]
        return values

    def get_property_at_step(self, step, varname, history_section='History') :
        """ Return the value of *varname* in the history section *history_section at step *step*."""
        if not 'ams' in self.rkfs: return
        main = self.rkfs['ams']
        as_block = self._values_stored_as_blocks(main, varname, history_section)
        if as_block :
            import numpy
            blocksize = main.read(history_section,'blockSize')
            iblock = int(numpy.ceil(step/blocksize))
            value = main.read(history_section,f"{varname}({iblock})")[(step%blocksize)-1]
        else :
            value = main.read(history_section,f"{varname}({step})")
        return value

    def _values_stored_as_blocks(self, main, varname, history_section) :
        """Determines wether the values of varname in a trajectory rkf file are stored in blocks"""
        nentries = main.read(history_section,'nEntries')
        as_block = False
        keylist = [var for sec,var in main if sec==history_section]
        if 'nBlocks' in keylist :
            if not f"{varname}({nentries})" in keylist :
                as_block = True
        return as_block

    def get_engine_results(self, engine=None):
        """Return a dictionary with contents of ``AMSResults`` section from an engine results ``.rkf`` file.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """
        return self._process_engine_results(lambda x: x.read_section('AMSResults'), engine)


    def get_engine_properties(self, engine=None):
        """Return a dictionary with all the entries from ``Properties`` section from an engine results ``.rkf`` file.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """

        def properties(kf):
            if not 'Properties' in kf:
                return {}
            # There are two *kinds* of "Properties" sections:
            # - ADF simply has a set of variables in the properties section
            # - DFTB and some other engines have a *scheme* for storing "Properties". See the fortran 'RKFileModule' (RKFile.f90) 
            #   for more details on this.
            ret = {}
            if ('Properties', 'nEntries') in kf:
                # This is a 'RKFileModule' properties section:
                n = kf.read('Properties', 'nEntries')
                for i in range(1, n+1):
                    tp = kf.read('Properties', 'Type({})'.format(i)).strip()
                    stp = kf.read('Properties', 'Subtype({})'.format(i)).strip()
                    val = kf.read('Properties', 'Value({})'.format(i))
                    key = stp if stp.endswith(tp) else ('{} {}'.format(stp, tp) if stp else tp)
                    ret[key] = val
            else:
                skeleton = kf.get_skeleton()
                for variable in skeleton['Properties']:
                    ret[variable] = kf.read('Properties', variable)
            return ret
        return self._process_engine_results(properties, engine)


    def get_energy(self, unit='au', engine=None):
        """Return final energy, expressed in *unit*.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """
        return self._process_engine_results(lambda x: x.read('AMSResults', 'Energy'), engine) * Units.conversion_ratio('au', unit)


    def get_gradients(self, energy_unit='au', dist_unit='au', engine=None):
        """Return the gradients of the final energy, expressed in *energy_unit* / *dist_unit*.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """
        return np.asarray(self._process_engine_results(lambda x: x.read('AMSResults', 'Gradients'), engine)).reshape(-1,3) * Units.conversion_ratio('au', energy_unit) / Units.conversion_ratio('au', dist_unit)


    def get_stresstensor(self, engine=None):
        """Return the final stress tensor, expressed in atomic units.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """
        return np.asarray(self._process_engine_results(lambda x: x.read('AMSResults', 'StressTensor'), engine)).reshape(len(self.get_input_molecule().lattice),-1)


    def get_hessian(self, engine=None):
        """Return the Hessian matrix, i.e. the second derivative of the total energy with respect to the nuclear coordinates, expressed in atomic units.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """
        return np.asarray(self._process_engine_results(lambda x: x.read('AMSResults', 'Hessian'), engine)).reshape(3*len(self.get_input_molecule()),-1)


    def get_elastictensor(self, engine=None):
        """Return the elastic tensor, expressed in atomic units.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """
        et_flat = np.asarray(self._process_engine_results(lambda x: x.read('AMSResults', 'ElasticTensor'), engine))
        num_latvec = len(self.get_input_molecule().lattice)
        if num_latvec == 1:
            return et_flat.reshape(1,1)
        elif num_latvec == 2:
            return et_flat.reshape(3,3)
        else:
            return et_flat.reshape(6,6)


    def get_frequencies(self, unit='cm^-1', engine=None):
        """Return a numpy array of vibrational frequencies, expressed in *unit*.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """
        freqs = self._process_engine_results(lambda x: x.read('Vibrations', 'Frequencies[cm-1]'), engine)
        freqs = np.array(freqs) if isinstance(freqs,list) else np.array([freqs])
        return freqs * Units.conversion_ratio('cm^-1', unit)


    def get_charges(self, engine=None):
        """Return the atomic charges, expressed in atomic units.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """
        return np.asarray(self._process_engine_results(lambda x: x.read('AMSResults', 'Charges'), engine))


    def get_dipolemoment(self, engine=None):
        """Return the electric dipole moment, expressed in atomic units.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """
        return np.asarray(self._process_engine_results(lambda x: x.read('AMSResults', 'DipoleMoment'), engine))


    def get_dipolegradients(self, engine=None):
        """Return the nuclear gradients of the electric dipole moment, expressed in atomic units. This is a (3*numAtoms x 3) matrix.
        """
        return np.asarray(self._process_engine_results(lambda x: x.read('AMSResults', 'DipoleGradients'), engine)).reshape(-1,3)


    def get_timings(self):
        """Return a dictionary with timing statistics of the job execution. Returned dictionary contains keys cpu, system and elapsed. The values are corresponding timings, expressed in seconds.
        """
        ret = {}
        try:
            # new AMS versions store timings on ams.rkf
            ret['elapsed'] = self.readrkf('General', 'ElapsedTime')
            ret['system'] = self.readrkf('General', 'SysTime')
            ret['cpu'] = self.readrkf('General', 'CPUTime')
        except:
            # fall back to reading output, was needed for old AMS versions
            cpu = self.grep_output('Total cpu time:')
            system = self.grep_output('Total system time:')
            elapsed = self.grep_output('Total elapsed time:')
            ret['elapsed'] = float(elapsed[0].split()[-1])
            ret['system'] = float(system[0].split()[-1])
            ret['cpu'] = float(cpu[0].split()[-1])

        return ret


    def get_youngmodulus(self, unit='au', engine=None):
        return self._process_engine_results(lambda x: x.read('AMSResults', 'YoungModulus'), engine) * Units.conversion_ratio('au', unit)

    def get_shearmodulus(self, unit='au', engine=None):
        return self._process_engine_results(lambda x: x.read('AMSResults', 'ShearModulus'), engine) * Units.conversion_ratio('au', unit)

    def get_bulkmodulus(self, unit='au', engine=None):
        return self._process_engine_results(lambda x: x.read('AMSResults', 'BulkModulus'), engine) * Units.conversion_ratio('au', unit)

    def get_pesscan_results(self, molecules=True):
        """
            For PESScan jobs, this functions extracts information about the scan coordinates and energies.

            molecules : bool
                Whether to return a Molecule at each PES point.

            Returns a dictionary, with the following keys:
            
            'RaveledScanCoords': a list of str with the names of the scan coordinates: ['sc_1_1','sc_1_2','sc_2_1','sc_2_2',...]

            'nRaveledScanCoords': the length of the previous list

            'RaveledUnits': a list of str with the units: ['bohr', 'bohr', 'radian', 'bohr', ...]

            'RaveledPESCoords': a nested list with the values of the scan coordinates: [[val_1_1_1,val_1_1_2,...],[val_1_2_1,val_1_2_2,...],[val_2_1_1,val_2_1_2,...],[val_2_2_1,val_2_2_2]]

            'ScanCoords': a nested list: [['sc_1_1','sc_1_2'],['sc_2_1','sc_2_2'],...]

            'nScanCoords': length of the previous list

            'Units': a nested list with the units: [['bohr', 'bohr'],['radian', 'bohr'], ...]

            'OrigScanCoords': a list of str in the newline-separated format stored on the .rkf file: ['sc_1_1\\nsc_1_2','sc_2_1\\nsc_2_2',...]

            'nPESPoints': int, number of PES points

            'PES': list of float, the energies at each PES point

            'Converged': list of bool, whether the geometry optimization at each PES point converged

            'Molecules': list of Molecule, the structure at each PES point. Only if the property molecules == True

            'HistoryInidices': list of int, the indices (1-based) in the History section which correspond to the Molecules and PES.

        """
        def tolist(x):
            if isinstance(x, list):
                return x
            else:
                return [x]

        nScanCoord = self.readrkf('PESScan', 'nScanCoord')
        nPoints = tolist(self.get_history_property('nPoints', 'PESScan'))

        pes = tolist(self.readrkf('PESScan', 'PES'))

        origscancoords = tolist(self.get_history_property('ScanCoord', history_section='PESScan'))
        # one scan coordinate may have several variables
        scancoords = [x.split('\n') for x in origscancoords]
        units = []
        pescoords = tolist(self.readrkf('PESScan', 'PESCoords'))
        pescoords = np.array(pescoords).reshape(-1,sum(len(x) for x in scancoords))
        pescoords = np.transpose(pescoords)
        w = []
        units = []
        count = 0
        for i in range(nScanCoord):
            units.append([])
            for j in range(len(scancoords[i])):
                if 'Distance' in scancoords[i][j]:
                    units[-1].append('bohr')
                else:
                    units[-1].append('radian')

        converged = tolist(self.readrkf('PESScan', 'GOConverged'))
        converged = [bool(x) for x in converged]
        historyindices = tolist(self.readrkf('PESScan', 'HistoryIndices'))

        ret = {}

        raveled_scancoords = [x[i] for x in scancoords for i in range(len(x))] # 1d list
        raveled_units = [x[i] for x in units for i in range(len(x))] # 1d list
        ret['RaveledScanCoords'] = raveled_scancoords
        ret['nRaveledScanCoords'] = len(raveled_scancoords)
        ret['ScanCoords'] = scancoords
        ret['nScanCoords'] = len(scancoords)
        ret['OrigScanCoords'] = origscancoords #newline delimiter for joint scan coordinates
        ret['RaveledPESCoords'] = pescoords.tolist()
        ret['Units'] = units
        ret['RaveledUnits'] = raveled_units
        ret['Converged'] = converged
        ret['PES'] = pes
        ret['nPESPoints'] = len(pes)
        ret['HistoryIndices'] = historyindices

        if molecules:
            mols = []
            for ind in historyindices:
                mols.append(self.get_history_molecule(ind))
            ret['Molecules'] = mols
        
        return ret

 
    def recreate_molecule(self):
        """Recreate the input molecule for the corresponding job based on files present in the job folder. This method is used by |load_external|.

        If ``ams.rkf`` is present in the job folder, extract data from the ``InputMolecule`` section.
        """
        if 'ams' in self.rkfs:
            return self.get_input_molecule()
        return None


    def recreate_settings(self):
        """Recreate the input |Settings| instance for the corresponding job based on files present in the job folder. This method is used by |load_external|.

        If ``ams.rkf`` is present in the job folder, extract user input and parse it back to a |Settings| instance using ``scm.input_parser`` module. Remove the ``system`` branch from that instance.
        """
        if 'ams' in self.rkfs:
            user_input = self.readrkf('General', 'user input')
            try:
                from scm.input_parser import InputParser
                with InputParser() as parser:
                    inp = parser.to_settings('ams', user_input)
            except:
                log('Failed to recreate input settings from {}'.format(self.rkfs['ams'].path, 5))
                return None
            s = Settings()
            s.input = inp
            if 'system' in s.input.ams:
                del s.input.ams.system
            s.soft_update(config.job)
            return s
        return None


    def ok(self):
        """Check if the execution of the associated :attr:`job` was successful or not.
        See :meth:`Job.ok<scm.plams.core.basejob.Job.ok>` for more information."""
        return self.job.ok()


    def get_errormsg(self):
        """Tries to get an an error message for a associated :attr:`job`. This method returns ``None`` if the associated job was successful.
        See :meth:`Job.get_errormsg<scm.plams.core.basejob.Job.errormsg>` for more information."""
        return self.job.get_errormsg()


    @property
    def name(self):
        """Retrun the :attr:`job.name` of the job associated with this results instance."""
        return self.job.name


    class EnergyLandscape:

        class State:
            def __init__(self, landscape, engfile, energy, mol, count, isTS, reactantsID=None, productsID=None):
                self._landscape = landscape
                self.engfile = engfile
                self.energy = energy
                self.molecule = mol
                self.count = count
                self.isTS = isTS
                self.reactantsID = reactantsID
                self.productsID = productsID

            @property
            def id(self):
                return self._landscape._states.index(self)+1

            @property
            def reactants(self):
                return self._landscape._states[self.reactantsID-1]

            @property
            def products(self):
                return self._landscape._states[self.productsID-1]

            def __str__(self):
                if self.isTS:
                    lines  = [f"State {self.id}: transition state @ {self.energy} Hartree (found {self.count} times, results on {self.engfile})"]
                    if self.reactantsID != None: lines += [f"|- Reactants: {self.reactants}"]
                    if self.productsID  != None: lines += [f"+- Products:  {self.products}"]
                else:
                    lines  = [f"State {self.id}: local minimum @ {self.energy} Hartree (found {self.count} times, results on {self.engfile})"]
                return "\n".join(lines)

        def __init__(self, results):
            sec = results.read_rkf_section("EnergyLandscape")

            # If there is only 1 state in the 'EnergyLandscape' section, some variables that are normally lists are insted be build-in types (e.g. a 'float' instead of a 'list of floats'). 
            # For convenience here we make sure that the following variables are always 'lists':
            for var in ['energies', 'counts', 'isTS', 'reactants', 'products']:
                if not isinstance(sec[var], list):
                    sec[var] = [sec[var]]

            nStates = sec['nStates']
            self._states = []

            for iState in range(nStates):
                energy  = sec['energies'][iState]
                resfile = os.path.splitext(sec['fileNames'].split('\0')[iState])[0]
                mol = results.get_molecule('Molecule',file=resfile)
                count = sec['counts'][iState]
                if not sec['isTS'][iState]:
                    self._states.append(AMSResults.EnergyLandscape.State(self, resfile, energy, mol, count, False))
                else:
                    reactantsID = sec['reactants'][iState] if sec['reactants'][iState] > 0 else None
                    productsID  = sec['products'][iState] if sec['products'][iState] > 0 else None
                    self._states.append(AMSResults.EnergyLandscape.State(self, resfile, energy, mol, count, True, reactantsID, productsID))

        @property
        def minima(self):
            return [s for s in self._states if not s.isTS ]

        @property
        def transition_states(self):
            return [s for s in self._states if s.isTS]

        def __str__(self):
            return 'All stationary points:\n'+\
                   '======================\n'+\
                   '\n'.join(str(s) for s in self._states)

        def __getitem__(self, i):
            return self._states[i-1]

        def __iter__(self):
            return iter(self._states)

        def __len__(self):
            return len(self._states)


    def get_energy_landscape(self):
        """Returns the energy landscape obtained from a PESExploration job run by the AMS driver. The energy landscape is a set of stationary PES points (local minima and transition states).

        The returned object is of the type ``AMSResults.EnergyLandscape`` and offers convenient access to the states' energies, geometries as well as the information on which transition states connect which minima.

        .. code-block:: python

            el = results.get_energy_landscape()
            print(el)

            for state in el:
                print(f"Energy = {state.energy}")
                print(f"Is transition state = {state.isTS}")
                print("Geometry:", state.molecule)
                if state.isTS:
                    print(f"Forward  barrier: {state.energy - state.reactants.energy}")
                    print(f"Backward barrier: {state.energy - state.products.energy}")
        """
        return AMSResults.EnergyLandscape(self)


    #=========================================================================


    def _access_rkf(self, func, file='ams'):
        """A skeleton method for accessing any of the ``.rkf`` files produced by AMS.

        The *file* argument should be the identifier of the file to read. It defaults to ``'ams'``. To access a file called ``something.rkf`` you need to call this function with ``file='something'``. If there exists only one engine results ``.rkf`` file, you can call this function with ``file='engine'`` to access this file.

        The *func* argument has to be a function to call on a chosen ``.rkf`` file. It should take one argument, an instance of |KFFile|.
        """
        #Try unique engine:
        if file == 'engine':
            names = self.engine_names()
            if len(names) == 1:
                return func(self.rkfs[names[0]])
            else:
                raise ValueError("You cannot use 'engine' as 'file' argument if the engine results file is not unique. Please use the real name of the file you wish to read")

        #Try:
        if file in self.rkfs:
            return func(self.rkfs[file])

        #Try harder:
        filename = file + '.rkf'
        self.refresh()
        if filename in self.files:
            self.rkfs[file] = KFFile(opj(self.job.path, filename))
            return func(self.rkfs[file])

        #Surrender:
        raise FileError('File {} not present in {}'.format(filename, self.job.path))


    def _process_engine_results(self, func, engine=None):
        """A generic method skeleton for processing any engine results ``.rkf`` file. *func* should be a function that takes one argument (an instance of |KFFile|) and returns arbitrary data.

        The *engine* argument should be the identifier of the file you wish to read. To access a file called ``something.rkf`` you need to call this function with ``engine='something'``. The *engine* argument can be omitted if there's only one engine results file in the job folder.
        """
        names = self.engine_names()
        if engine is not None:
            if engine in names:
                return func(self.rkfs[engine])
            else:
                raise FileError('File {}.rkf not present in {}'.format(engine, self.job.path))
        else:
            if len(names) == 1:
                return func(self.rkfs[names[0]])
            elif len(names) == 0:
                raise FileError('There is no engine .rkf present in {}'.format(self.job.path))
            else:
                raise ValueError("You need to specify the 'engine' argument when there are multiple engine result files present in the job folder")


#===========================================================================
#===========================================================================
#===========================================================================


class AMSJob(SingleJob):
    """A class representing a single computation with AMS driver. The corresponding results type is |AMSResults|.
    """
    _result_type = AMSResults
    _command = 'ams'


    def get_input(self):
        """Generate the input file. This method is just a wrapper around :meth:`_serialize_input`.

        Each instance of |AMSJob| or |AMSResults| present as a value in ``settings.input`` branch is replaced with an absolute path to ``ams.rkf`` file of that job.

        If you need to use a path to some engine specific ``.rkf`` file rather than the main ``ams.rkf`` file, you can to it by supplying a tuple ``(x, name)`` where ``x`` is an instance of |AMSJob| or |AMSResults| and ``name`` is a string with the name of the ``.rkf`` file you want. For example, ``(myjob, 'dftb')`` will transform to the absolute path to ``dftb.rkf`` file in ``myjob``'s folder, if such a file is present.

        Instances of |KFFile| are replaced with absolute paths to corresponding files.
        """

        special = {
            AMSJob: lambda x: x.results.rkfpath(),
            AMSResults: lambda x: x.rkfpath(),
            KFFile: lambda x: x.path,
            tuple: lambda x: AMSJob._tuple2rkf(x)
        }
        return self._serialize_input(special)


    def get_runscript(self):
        """Generate the runscript. Returned string is of the form::

            unset AMS_SWITCH_LOGFILE_AND_STDOUT
            AMS_JOBNAME=jobname AMS_RESULTSDIR=. $AMSBIN/ams [-n nproc] <jobname.in [>jobname.out]

        ``-n`` flag is added if ``settings.runscript.nproc`` exists. ``[>jobname.out]`` is used based on ``settings.runscript.stdout_redirect``. If ``settings.runscript.preamble_lines`` exists, those lines will be added to the runscript verbatim before the execution of AMS.
        """
        ret  = 'unset AMS_SWITCH_LOGFILE_AND_STDOUT\n'
        if 'preamble_lines' in self.settings.runscript:
            for line in self.settings.runscript.preamble_lines:
                ret += f'{line}\n'
        ret += 'AMS_JOBNAME="{}" AMS_RESULTSDIR=. $AMSBIN/ams'.format(self.name)
        if 'nproc' in self.settings.runscript:
            ret += ' -n {}'.format(self.settings.runscript.nproc)
        ret += ' <"{}"'.format(self._filename('inp'))
        if self.settings.runscript.stdout_redirect:
            ret += ' >"{}"'.format(self._filename('out'))
        ret += '\n\n'
        return AMSJob._slurm_env(self.settings) + ret


    def check(self):
        """Check if ``termination status`` variable from ``General`` section of main KF file equals ``NORMAL TERMINATION``."""
        try:
            status = self.results.readrkf('General', 'termination status')
        except:
            return False
        if 'NORMAL TERMINATION' in status:
            if 'errors' in status:
                log('Job {} reported errors. Please check the the output'.format(self._full_name()), 1)
                return False
            if 'warnings' in status:
                log('Job {} reported warnings. Please check the the output'.format(self._full_name()), 1)
            return True
        return False


    def get_errormsg(self):
        """Tries to get an an error message for a failed job. This method returns ``None`` for successful jobs."""
        if self.ok():
            return None
        else:
            # Something went wrong. The first place to check is the termination status on the ams.rkf.
            # If the AMS driver stopped with a known error (called StopIt in the Fortran code), the error will be in there.
            try:
                msg = self.results.readrkf('General','termination status')
                if msg == 'NORMAL TERMINATION with errors':
                    # Apparently this wasn't a hard stop in the middle of the job.
                    # Let's look for the last error in the logfile ...
                    msg = self.results.grep_file('ams.log', 'ERROR: ')[-1].partition('ERROR: ')[2]
            except:
                msg = 'Could not determine error message. Please check the output manually.'
            return msg



    def hash_input(self):
        """Calculate the hash of the input file.

        All instances of |AMSJob| or |AMSResults| present as values in ``settings.input`` branch are replaced with hashes of corresponding job's inputs. Instances of |KFFile| are replaced with absolute paths to corresponding files.
        """
        special = {
            AMSJob: lambda x: x.hash_input(),
            AMSResults: lambda x: x.job.hash_input(),
            KFFile: lambda x: x.path,
            tuple: lambda x: AMSJob._tuple2rkf(x)
        }
        return sha256(self._serialize_input(special))


    #=========================================================================


    def _serialize_input(self, special):
        """Transform the contents of ``settings.input`` branch into string with blocks, keys and values.

        First, the contents of ``settings.input`` are extended with entries returned by :meth:`_serialize_molecule`. Then the contents of ``settings.input.ams`` are used to generate AMS text input. Finally, every other (than ``ams``) entry in ``settings.input`` is used to generate engine specific input.

        Special values can be indicated with *special* argument, which should be a dictionary having types of objects as keys and functions translating these types to strings as values.
        """

        def unspec(value):
            """Check if *value* is one of a special types and convert it to string if it is."""
            for spec_type in special:
                if isinstance(value, spec_type):
                    return special[spec_type](value)
            return value

        def serialize(key, value, indent, end='End'):
            """Given a *key* and its corresponding *value* from the |Settings| instance produce a snippet of the input file representing this pair.

            If the value is a nested |Settings| instance, use recursive calls to build the snippet for the entire block. Indent the result with *indent* spaces.
            """
            ret = ''
            if isinstance(value, Settings):
                ret += ' '*indent + key
                if '_h' in value:
                    ret += ' ' + unspec(value['_h'])
                ret += '\n'

                i = 1
                while ('_'+str(i)) in value:
                    ret += serialize('', value['_'+str(i)], indent+2)
                    i += 1

                for el in value:
                    if not el.startswith('_'):
                        if key.lower().startswith('engine') and el.lower() == 'input':
                            ret += serialize(el, value[el], indent+2, 'EndInput')
                        # REB: For the hybrid engine. How todeal with the space in el (Engine DFTB)? Replace by underscore?
                        elif key.lower().startswith('engine') and el.lower().startswith('engine') :
                            engine = ' '.join(el.split('_'))
                            ret += serialize(engine, value[el], indent+2, end='EndEngine') + '\n'
                        else:
                            ret += serialize(el, value[el], indent+2)

                if key.lower() == 'input': end='endinput'
                ret += ' '*indent + end+'\n'

            elif isinstance(value, list):
                for el in value:
                    ret += serialize(key, el, indent)
            elif value == '' or value is True:
                ret += ' '*indent + key + '\n'
            elif value is False or value is None:
                pass
            else:
                ret += ' '*indent + key + ' ' + str(unspec(value)) + '\n'
            return ret

        fullinput = self.settings.input.copy()

        #prepare contents of 'system' block(s)
        more_systems = self._serialize_molecule()
        if more_systems:
            if 'system' in fullinput.ams:
                #nonempty system block was already present in input.ams
                system = fullinput.ams.system
                system_list = system if isinstance(system, list) else [system]

                system_list_set = Settings({(s._h if '_h' in s else ''):s   for s in system_list})
                more_systems_set = Settings({(s._h if '_h' in s else ''):s   for s in more_systems})

                system_list_set += more_systems_set
                system_list = list(system_list_set.values())
                system = system_list[0] if len(system_list) == 1 else system_list
                fullinput.ams.system = system

            else:
                fullinput.ams.system = more_systems[0] if len(more_systems) == 1 else more_systems

        txtinp = ''
        ams = fullinput.find_case('ams')

        #contents of the 'ams' block (AMS input) go first
        for item in fullinput[ams]:
            txtinp += serialize(item, fullinput[ams][item], 0) + '\n'

        #and then engines
        for engine in fullinput:
            if engine != ams:
                txtinp += serialize('Engine '+engine, fullinput[engine], 0, end='EndEngine') + '\n'

        return txtinp


    def _serialize_molecule(self):
        """Return a list of |Settings| instances containing the information about one or more |Molecule| instances stored in the ``molecule`` attribute.

        Molecular charge is taken from ``molecule.properties.charge``, if present. Additional, atom-specific information to be put in ``atoms`` block after XYZ coordinates can be supplied with ``atom.properties.suffix``.

        If the ``molecule`` attribute is a dictionary, the returned list is of the same length as the size of the dictionary. Keys from the dictionary are used as headers of returned ``system`` blocks.
        """

        if self.molecule is None:
            return Settings()

        moldict = {}
        if isinstance(self.molecule, Molecule):
            moldict = {'':self.molecule}
        elif isinstance(self.molecule, dict):
            moldict = self.molecule
        else:
            raise JobError("Incorrect 'molecule' attribute of job {}. 'molecule' should be a Molecule, a dictionary or None, and not {}".format(self._full_name(), type(self.molecule)))

        ret = []
        for name, molecule in moldict.items():
            newsystem = Settings()
            if name:
                newsystem._h = name

            if len(molecule.lattice) in [1,2] and molecule.align_lattice():
                log("The lattice of {} Molecule supplied for job {} did not follow the convention required by AMS. I rotated the whole system for you. You're welcome".format(name if name else 'main', self._full_name()), 3)

            newsystem.Atoms._1 = [atom.str(symbol=self._atom_symbol(atom), space=18, decimal=10,
                    suffix=(atom.properties.suffix if 'suffix' in atom.properties else '')) for atom in molecule]

            if molecule.lattice:
                newsystem.Lattice._1 = ['{:16.10f} {:16.10f} {:16.10f}'.format(*vec) for vec in molecule.lattice]

            if len(molecule.bonds)>0:
                newsystem.BondOrders._1 = ['{} {} {}'.format(molecule.index(b.atom1), molecule.index(b.atom2), b.order) for b in molecule.bonds]

            if 'charge' in molecule.properties:
                newsystem.Charge = molecule.properties.charge

            ret.append(newsystem)

        return ret


    #=========================================================================

    def get_task(self):
        """ Returns the AMS Task from the job's settings. If it does not exist, returns None. """
        if isinstance(self.settings, Settings) and 'input' in self.settings and 'ams' in self.settings.input and 'task' in self.settings.input.ams:
            return self.settings.input.ams.task
        return None


    @staticmethod
    def _atom_symbol(atom):
        """Return the atomic symbol of *atom*. Ensure proper formatting for AMSuite input taking into account ``ghost`` and ``name`` entries in ``properties`` of *atom*."""
        smb = atom.symbol if atom.atnum > 0 else ''  #Dummy atom should have '' instead of 'Xx'
        if 'ghost' in atom.properties and atom.properties.ghost:
            smb = ('Gh.'+smb).rstrip('.')
        if 'name' in atom.properties:
            smb = (smb+'.'+str(atom.properties.name)).lstrip('.')
        return smb


    @staticmethod
    def _tuple2rkf(arg):
        """Transform a pair ``(x, name)`` where ``x`` is an instance of |AMSJob| or |AMSResults| and ``name`` is a name of ``.rkf`` file (``ams`` or engine) to an absolute path to that ``.rkf`` file."""
        if len(arg) == 2 and isinstance(arg[1], str):
            if isinstance(arg[0], AMSJob):
                return arg[0].results.rkfpath(arg[1])
            if isinstance(arg[0], AMSResults):
                return arg[0].rkfpath(arg[1])
        return str(arg)


    @classmethod
    def load_external(cls, path, settings=None, molecule=None, finalize=False):
        """Load an external job from *path*.

        In this context an "external job" is an execution of some external binary that was not managed by PLAMS, and hence does not have a ``.dill`` file. It can also be used in situations where the execution was started with PLAMS, but the Python process was terminated before the execution finished, resulting in steps 9-12 of :ref:`job-life-cycle` not happening.

        All the files produced by your computation should be placed in one folder and *path* should be the path to this folder or a file in this folder. The name of the folder is used as a job name. Input, output, error and runscript files, if present, should have names defined in ``_filenames`` class attribute (usually ``[jobname].in``, ``[jobname].out``, ``[jobname].err`` and ``[jobname].run``). It is not required to supply all these files, but in most cases one would like to use at least the output file, in order to use methods like :meth:`~scm.plams.core.results.Results.grep_output` or :meth:`~scm.plams.core.results.Results.get_output_chunk`. If *path* is an instance of an AMSJob, that instance is returned.

        This method is a class method, so it is called via class object and it returns an instance of that class::

            >>> a = AMSJob.load_external(path='some/path/jobname')
            >>> type(a)
            scm.plams.interfaces.adfsuite.ams.AMSJob

        You can supply |Settings| and |Molecule| instances as *settings* and *molecule* parameters, they will end up attached to the returned job instance. If you don't do this, PLAMS will try to recreate them automatically using methods :meth:`~scm.plams.core.results.Results.recreate_settings` and :meth:`~scm.plams.core.results.Results.recreate_molecule` of the corresponding |Results| subclass. If no |Settings| instance is obtained in either way, the defaults from ``config.job`` are copied.

        You can set the *finalize* parameter to ``True`` if you wish to run the whole :meth:`~Job._finalize` on the newly created job. In that case PLAMS will perform the usual :meth:`~Job.check` to determine the job status (*successful* or *failed*), followed by cleaning of the job folder (|cleaning|), |postrun| and pickling (|pickling|). If *finalize* is ``False``, the status of the returned job is *copied*.

        This method can also be used to convert a finished VASP job to an AMSJob. If you supply the path to a folder containing OUTCAR, then a subdirectory will be created in this folder called AMSJob. In the AMSJob subdirectory, two files will be created: ams.rkf and vasp.rkf, that contain some of the results from the VASP calculation. If the AMSJob subdirectory already exists, the existing ams.rkf and vasp.rkf files will be reused. NOTE: the purpose of loading VASP data this way is to let you call for example job.results.get_energy() etc., not to run new VASP calculations!
        """
        if isinstance(path, cls):
            return path

        preferred_name = None
        # first check if path is a VASP output
        # in which case call cls._vasp_to_ams, which will return a NEW path 
        # containing ams.rkf and vasp.rkf (the .rkf files will be created if they do not exist)
        if os.path.isdir(path):
            preferred_name = os.path.basename(os.path.abspath(path))
            if not os.path.exists(os.path.join(path,'ams.rkf')) and os.path.exists(os.path.join(path,'OUTCAR')):
                path = vasp_output_to_ams(path, overwrite=False)
        elif os.path.exists(path) and os.path.basename(path) == 'OUTCAR':
            preferred_name = os.path.basename(os.path.dirname(os.path.abspath(path)))
            path = vasp_output_to_ams(os.path.dirname(path), overwrite=False)

        if not os.path.isdir(path):
            if os.path.exists(path):
                path = os.path.dirname(path)
            elif os.path.isdir(path+'.results'):
                path=path+'.results'
            elif os.path.isdir(path+'results'):
                path=path+'results'
            else:
                raise FileError('Path {} does not exist, cannot load from it.'.format(path))

        job = super(AMSJob, cls).load_external( path, settings, molecule, finalize)

        if preferred_name is not None:
            job.name = preferred_name

        if job.name.endswith('.results') and len(job.name) > 8:
            job.name = job.name[:-8]

        return job


    @classmethod
    def from_inputfile(cls, filename: str, heredoc_delimit: str = 'eor', **kwargs) -> 'AMSJob':
        """Construct an :class:`AMSJob` instance from an ADF inputfile or runfile.

        If a runscript is provide than this method will attempt to extract the input file based
        on the heredoc delimiter (see *heredoc_delimit*).

        """
        try:
            from scm.input_parser import InputParser
        except ImportError:  # Try to load the parser from $AMSHOME/scripting
            with UpdateSysPath():
                from scm.input_parser import InputParser

        s = Settings()
        with open(filename, 'r') as f:
            inp_file = parse_heredoc(f.read(), heredoc_delimit)

        with InputParser() as parser:
            s.input = parser.to_settings(cls._command, inp_file)
        if not s.input:
            raise JobError(f"from_inputfile: failed to parse '{filename}'")

        # Extract a molecule from the input settings
        mol = cls.settings_to_mol(s)

        # Create and return the Job instance
        if mol is not None:
            return cls(molecule=mol, settings=s, **kwargs)
        else:
            s.ignore_molecule = True
            return cls(settings=s, **kwargs)


    @staticmethod
    def settings_to_mol(s: Settings) -> dict:
        """Pop the `s.input.ams.system` block from a settings instance and convert it into a dictionary of molecules.

        The provided settings should be in the same style as the ones produced by the SCM input parser.
        Dictionary keys are taken from the header of each system block.
        The existing `s.input.ams.system` block is removed in the process, assuming it was present in the first place.

        """
        def read_mol(settings_block: Settings) -> Molecule:
            """Retrieve single molecule from a single `s.input.ams.system` block."""
            if 'geometryfile' in settings_block:
                mol = Molecule(settings_block.geometryfile)
            else:
                mol = Molecule()
                for atom in settings_block.atoms._1:
                    # Extract arguments for Atom()
                    symbol, x, y, z, *comment = atom.split(maxsplit=4)
                    kwargs = {} if not comment else {'suffix': comment[0]}
                    coords = float(x), float(y), float(z)

                    try:
                        at = Atom(symbol=symbol, coords=coords, **kwargs)
                    except PTError:  # It's either a ghost atom and/or an atom with a custom name
                        if symbol.startswith('Gh.'):  # Ghost atom
                            kwargs['ghost'], symbol = symbol.split('.', maxsplit=1)
                        if '.' in symbol:  # Atom with a custom name
                            symbol, kwargs['name'] = symbol.split('.', maxsplit=1)
                        at = Atom(symbol=symbol, coords=coords, **kwargs)

                    mol.add_atom(at)

                # Set the lattice vector if applicable
                if settings_block.lattice._1:
                    mol.lattice = [tuple(float(j) for j in i.split()) for i in settings_block.lattice._1]

            # Add bonds
            for bond in settings_block.bondorders._1:
                _at1, _at2, _order = bond.split()
                at1, at2, order = mol[int(_at1)], mol[int(_at2)], float(_order)
                mol.add_bond(at1, at2, order)

            # Set the molecular charge
            if settings_block.charge:
                mol.properties.charge = float(settings_block.charge)

            mol.properties.name = str(settings_block._h)
            return mol

        # Raises a KeyError if the `system` key is absent
        with s.suppress_missing():
            try:
                settings_list = s.input.ams.pop('system')
            except KeyError:  # The block s.input.ams.system is absent
                return None

        # Create a new dictionary with system headers as keys and molecules as values
        moldict = {}
        for settings_block in settings_list:
            key = str(settings_block._h) if ('_h' in settings_block) else '' # Empty string used as default system name.
            if key in moldict:
                raise KeyError(f"Duplicate system headers found in s.input.ams.system: {repr(key)}")
            moldict[key] = read_mol(settings_block)

        return moldict


    @staticmethod
    def _slurm_env(settings):
        """Produce a string with environment variables declaration needed for running AMS on a SLURM managed system.

        If the key ``slurm`` is present in ``settings.run`` and it's ``True`` the returned string is of the form:

        ``export SCM_MPIOPTIONS="-n X -N Y\n"

        ``X`` is taken from ``settings.run.cores`` (if not present, falls back to ``settings.runscript.nproc``).
        ``Y`` is taken from ``settings.run.nodes``
        """

        if 'slurm' in settings.run and settings.run.slurm is True:
            slurmenv = ''
            if 'cores' in settings.run:
                slurmenv += f'-n {settings.run.cores} '
            elif 'nproc' in settings.runscript:
                slurmenv += f'-n {settings.runscript.nproc} '
            if 'nodes' in settings.run:
                slurmenv += f'-N {settings.run.nodes} '
            if slurmenv:
                return f'export SCM_MPIOPTIONS="{slurmenv}"\n'
        return ''




from .units import Units
from ..trajectories.rkffile import RKFTrajectoryFile
from ..core.private import saferun
from .kftools import KFFile
import os

__all__ = ['traj_to_rkf', 'vasp_output_to_ams']

def traj_to_rkf(trajfile,  rkftrajectoryfile):
    """
        Convert ase .traj file to .rkf file. NOTE: The order of atoms (or the number of atoms) cannot change between frames!

        trajfile : str
            path to a .traj file
        rkftrajectoryfile : str 
            path to the output .rkf file (will be created)

        Returns : 2-tuple (coords, cell)
            The final coordinates and cell in angstrom
    """
    from ase.io import read, Trajectory
    traj = Trajectory(trajfile)
    rkfout = RKFTrajectoryFile(rkftrajectoryfile, mode='wb')
    rkfout.store_historydata()
    rkfout.store_mddata()

    # Store the units properly
    mdunits = {}
    mdunits['PotentialEnergy'] = 'hartree'
    mdunits['KineticEnergy'] = 'hartree'
    mdunits['TotalEnergy'] = 'hartree'
    rkfout._set_mdunits(mdunits)

    energy_converter = Units.convert(1.0, 'eV', 'hartree')
    gradients_converter = Units.convert(1.0, 'eV/angstrom', 'hartree/bohr')
    stress_converter = Units.convert(1.0, 'eV/angstrom^3', 'hartree/bohr^3')

    coords, cell = None, None
    try:
        for i,atoms in enumerate(traj):
            if i == 0:
                rkfout.set_elements(atoms.get_chemical_symbols())
            coords = atoms.get_positions() # angstrom
            cell = atoms.get_cell()
            try:
                gradients = -atoms.get_forces()*gradients_converter
            except:
                gradients = None
            try:
                stresstensor = atoms.get_stress(voigt=False)*stress_converter
            except:
                stresstensor = None
            mddata = {}
            try:
                energy = atoms.get_potential_energy()*energy_converter
                mddata['PotentialEnergy'] = energy
            except:
                energy = None

            try:
                kinetic_energy = atoms.get_kinetic_energy()*energy_converter
                mddata['KineticEnergy'] = kinetic_energy
            except:
                kinetic_energy = None

            if 'PotentialEnergy' in mddata and 'KineticEnergy' in mddata:
                mddata['TotalEnergy'] = mddata['PotentialEnergy'] + mddata['KineticEnergy']
            
            if len(mddata) == 0:
                mddata = None

            # Create a historydata dictionary, to go into the History section
            historydata = {}
            if gradients is not None :
                historydata['Gradients'] = gradients
            if stresstensor is not None :
                historydata['StressTensor'] = stresstensor
            if len(historydata) == 0 :
                historydata = None

            rkfout.write_next(coords=coords, cell=cell, historydata=historydata, mddata=mddata)

    finally:
        rkfout.close()


    # the below is needed to be able to load the .rkf file with AMSJob.load_external()
    kf = KFFile(rkftrajectoryfile)
    kf['EngineResults%nEntries'] = 0
    kf['General%user input'] = '\xFF'.join(['Engine External','EndEngine'])

    return coords, cell

def vasp_output_to_ams(vasp_folder, wdir=None, overwrite=False, write_engine_rkf=True):
    """ 
        Converts VASP output (OUTCAR, ...) to AMS output (ams.rkf, vasp.rkf)

        Returns: a string containing the directory where ams.rkf was written

        vasp_folder : str
            path to a directory with an OUTCAR, INCAR, POTCAR etc. files

        wdir : str or None
            directory in which to write the ams.rkf and vasp.rkf files
            If None, a subdirectory "AMSJob" of vasp_folder will be created

        overwrite : bool
            if False, first check if wdir already contains ams.rkf and vasp.rkf, in which case do nothing
            if True, overwrite if exists

        write_engine_rkf : bool
            If True, also write vasp.rkf alongside ams.rkf. The vasp.rkf file will only contain an AMSResults section (energy, gradients, stress tensor). It will not contain the DOS or the band structure.
    """
    if not os.path.isdir(vasp_folder):
        raise ValueError('Directory {} does not exist'.format(vasp_folder))

    outcar = os.path.join(vasp_folder, 'OUTCAR')
    if not os.path.exists(outcar):
        raise ValueError('File {} does not exist, should be an OUTCAR file.'.format(outcar))

    if wdir is None:
        wdir = os.path.join(os.path.dirname(outcar),'AMSJob')
        os.makedirs(wdir, exist_ok=True)

    # exit early if ams.rkf already exists
    if os.path.exists(os.path.join(wdir, 'ams.rkf')) and not overwrite:
        return wdir

    # convert OUTCAR to a .traj file inside wdir
    trajfile = os.path.join(wdir,'outcar.traj')
    if os.path.exists(trajfile):
        os.remove(trajfile)
    cmd = [os.path.join(os.environ.get('AMSBIN'),'amspython'),'-m','ase','convert',outcar,trajfile]
    saferun(cmd)
    if not os.path.exists(os.path.join(wdir,'outcar.traj')):
        raise RuntimeError("Couldn't write outcar.traj in {}".format(wdir))

    # remove the target files first if overwrite
    kffile = os.path.join(wdir, 'ams.rkf')
    enginefile = os.path.join(wdir, 'vasp.rkf')
    if os.path.exists(kffile): 
        if overwrite:
            os.remove(kffile)
        else:
            raise RuntimeError("{} already exists, specify overwrite=True to overwrite".format(kffile))
    if os.path.exists(enginefile): 
        if overwrite:
            os.remove(enginefile)
        else:
            raise RuntimeError("{} already exists, specify overwrite=True to overwrite".format(enginefile))

    # convert the .traj file to ams.rkf
    coords, cell = traj_to_rkf(trajfile, kffile)

    # add extra info to the kffile
    kf = KFFile(kffile, autosave=False)
    if write_engine_rkf:
        enginerkf = KFFile(enginefile, autosave=False)
    try:
        kf['EngineResults%nEntries'] = 1
        kf['EngineResults%Title(1)'] = 'vasp'
        kf['EngineResults%Description(1)'] = 'Standalone VASP run. Data from {}'.format(os.path.abspath(outcar))
        kf['EngineResults%Files(1)'] = 'vasp.rkf'
        kf['General%user input'] = '!VASP'

        # read the INCAR
        incarfile = os.path.join(vasp_folder, 'INCAR')
        userinput = ['!VASP','Engine External', '  Free', '  !INCAR']
        if os.path.exists(incarfile):
            with open(incarfile) as incar:
                for line in incar:
                    line = line.split('!')[0]
                    line = line.split('#')[0]
                    line = line.strip()
                    if line.lower().startswith('end'): # "End" is reserved to end the block
                        line = '!'+line
                    if len(line) > 0:
                        userinput.append('    '+line)
            userinput.append('  !EndINCAR')
        userinput.append('  End') #end of the Free block
        userinput.append('EndEngine')
        kf['General%user input'] = '\xFF'.join(userinput)

        # write engine.rkf
        # copy General, Molecule, InputMolecule from ams.rkf
        if write_engine_rkf:
            for sec in ['General','Molecule','InputMolecule']:
                secdict = kf.read_section(sec)
                for k, v in secdict.items():
                    enginerkf[sec+'%'+k] = v
            nEntries = kf['History%nEntries']
            suffix='({})'.format(nEntries)
            if ('History', 'Energy'+suffix) in kf: enginerkf['AMSResults%Energy'] = kf['History%Energy'+suffix]
            if ('History', 'Gradients'+suffix) in kf: enginerkf['AMSResults%Gradients'] = kf['History%Gradients'+suffix]
            if ('History', 'StressTensor'+suffix) in kf: enginerkf['AMSResults%StressTensor'] = kf['History%StressTensor'+suffix]

    finally:
        kf.save()
        if write_engine_rkf:
            enginerkf.save()

    if os.path.exists(trajfile):
        os.remove(trajfile)

    return wdir




from .scmjob import SCMJob, SCMResults
from ...core.errors import ResultsError
from ...core.functions import log
from ...tools.units import Units

__all__ = ['DFTBJob', 'DFTBResults']



class DFTBResults(SCMResults):
    _kfext = '.rkf'
    _rename_map = {'dftb.rkf':'$JN'+_kfext}


    def get_main_molecule(self):
        """get_main_molecule()
        Return a |Molecule| instance based on the ``Molecule`` section in the main KF file (``$JN.rkf``).

        For runs with multiple geometries (geometry optimization, transition state search, molecular dynamics) this is the **final** geometry. In such a case, to access the initial (or any intermediate) coordinates please extract them from section ``History``, variable ``Coords(i)`` or use :meth:`get_input_molecule`. Mind the fact that all coordinates written by DFTB to ``.rkf`` file are in bohr::

            mol = results.get_molecule(section='History', variable='Coords(1)', unit='bohr')
        """
        ret = self.get_molecule(section='Molecule', variable='Coords', unit='bohr')
        ret.properties.charge = self.readkf('Molecule', 'Charge')
        if ('Molecule', 'LatticeVectors') in self._kf:
            lattice = self.readkf('Molecule', 'LatticeVectors')
            lattice = Units.convert(lattice, 'bohr', 'angstrom')
            ret.lattice = [tuple(lattice[i:i+3]) for i in range(0,len(lattice),3)]
        return ret


    def get_input_molecule(self):
        """get_input_molecule()
        Return a |Molecule| instance with initial coordinates.

        All data used by this method is taken from ``$JN.rkf`` file. The ``molecule`` attribute of the corresponding job is ignored.
        """
        if ('History', 'nEntries') in self._kf:
            return self.get_molecule(section='History', variable='Coords(1)', unit='bohr')
        return self.get_main_molecule()


    def get_history_molecules(self):
        """get_history_molecules()
        Return a list of |Molecule| instances with all Structures from the ``History`` section of the ``.rkf`` file.
        If the structures have lattice information, it is written into the corresponding |Molecule| instance.

        All data used by this method is taken from ``$JN.rkf`` file. If no ``History`` section is available an empty list is returned and a level 5 log entry appears.
        """
        if not ('History', 'nEntries') in self._kf:
            log('No History section found in rkf during get_history_molecules, returning an empty list.',level=5)
            return []

        nEntries = self.readkf('History', 'nEntries')
        #list of molecules to return
        history = []
        #iterate over entries
        for i in range(1,nEntries+1):
            mol = self.get_molecule('History','Coords('+str(i)+')')
            #read lattice if there is any
            if ('History','LatticeVectors('+str(i)+')') in self._kf:
                #LatticeVectors has to be length dimensions*3, so we can do this
                lattice = Units.convert(self.readkf('History','LatticeVectors('+str(i)+')'), 'bohr', 'angstrom')
                mol.lattice = [tuple(lattice[j:j+3]) for j in range(0,len(lattice),3)]
            history.append(mol)
        return history


    def get_energy(self, unit='au'):
        """get_energy(unit='au')
        Return DFTB final energy, expressed in *unit*.
        """
        prop = self.get_properties()
        if 'DFTB Final Energy' in prop:
            return Units.convert(prop['DFTB Final Energy'], 'au', unit)
        raise ResultsError("'DFTB Final Energy' not present in 'Properties' section of {}".format(self._kfpath()))


    def get_dipole_vector(self, unit='au'):
        """get_dipole_vector(unit='au')
        Return the dipole vector, expressed in *unit*.
        """
        prop = self.get_properties()
        if 'Electric Dipole Moment' in prop:
            return Units.convert(prop['Electric Dipole Moment'], 'au', unit)
        raise ResultsError("'Electric Dipole Moment' not present in 'Properties' section of {}".format(self._kfpath()))


    def get_gradients(self):
        """get_gradients()
        Return the list of atomic gradients, expressed in atomic units. Returned value is a list of vectors (lists of lenght 3).
        """
        prop = self.get_properties()
        if 'Generic Gradient' in prop:
            grad = prop['Generic Gradient']
            return [grad[i:i+3] for i in range(0,len(grad),3)]
        raise ResultsError("'Generic Gradient' not present in 'Properties' section of {}".format(self._kfpath()))


    def _int2inp(self):
        """_int2inp()
        In DFTB the internal order is always the same as the input order. Return an identity permutation of length equal to the number of atoms.
        """
        return list(range(1, 1+self.readkf('Molecule', 'nAtoms')))


    def _atomic_numbers_input_order(self):
        """_atomic_numbers_input_order()
        Return a list of atomic numbers, in the input order.
        """
        return self.readkf('Molecule', 'AtomicNumbers')



class DFTBJob(SCMJob):
    _result_type = DFTBResults
    _command = 'dftb'
    _top = ['units', 'task']
    _subblock_end = 'end'


    def _serialize_mol(self):
        s = self.settings.input

        if len(self.molecule.lattice) in [1,2] and self.molecule.align_lattice():
            log("The lattice supplied for job {} did not follow the convention required by DFTB. I rotated the whole system for you. You're welcome".format(self._full_name()), 3)

        for i,atom in enumerate(self.molecule):
            s.system.atoms['_'+str(i+1)] = atom.str(symbol=self._atom_symbol(atom), space=18, decimal=10)

        if self.molecule.lattice:
            for i,vec in enumerate(self.molecule.lattice):
                s.system.lattice['_'+str(i+1)] = '{:16.10f} {:16.10f} {:16.10f}'.format(*vec)


    def _remove_mol(self):
        s = self.settings.input
        if 'system' in s:
            if 'atoms' in s[system]:
                del s.system.atoms
            if 'lattice' in s.system:
                del s.system.lattice



from setuptools import setup, find_packages
from os.path import join as opj

packages = ['scm.plams'] + ['scm.plams.'+i for i in find_packages('.')]

description = "PLAMS is a library providing powerful, flexible and easily extendable Python interface to molecular modeling programs. It takes care of input preparation, job execution, file management and output data extraction. It helps with building advanced data workflows that can be executed in parallel, either locally or by submitting to a resource manager queue.\n\nPlease check the project's GitHub page for more information: https://github.com/SCM-NV/PLAMS \n\nPLAMS is an Open Source project supported by `Software for Chemistry & Materials B.V. <https://www.scm.com>`_"

setup(
    name             = 'plams',
    version          = '1.5.1',
    author           = 'Michał Handzlik',
    author_email     = 'handzlik@scm.com',
    url              = 'https://www.scm.com/doc/plams/',
    download_url     = 'https://github.com/SCM-NV/PLAMS/zipball/release',
    license          = 'LGPLv3',
    description      = 'Python Library for Automating Molecular Simulations',
    long_description = description,
    classifiers      = [
            'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Science/Research',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3.6',
            'Topic :: Scientific/Engineering :: Chemistry',
            'Topic :: Scientific/Engineering :: Physics',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords         = ['molecular modeling', 'computational chemistry', 'workflow', 'python interface'],
    python_requires  = '>=3.6',
    install_requires = ['dill>=0.2.4', 'numpy'],
    packages         = packages,
    package_dir      = {'scm.plams': '.'},
    package_data     = {'scm.plams' : ['plams_defaults', 'examples/*']},
    scripts          = [opj('scripts','plams')]
)



#!/usr/bin/env python

import numpy

from ..mol.molecule import Molecule, Bond
from ..tools.periodic_table import PeriodicTable
from ..tools.kftools import KFFile
from ..tools.units import Units
#from ..interfaces.adfsuite.ams import AMSResults
from ..core.errors import PlamsError
from .trajectoryfile import TrajectoryFile

__all__ = ['RKFTrajectoryFile']

bohr_to_angstrom = Units.conversion_ratio('bohr','angstrom')

class RKFTrajectoryFile (TrajectoryFile) :
        """
        Class representing an RKF file containing a molecular trajectory

        An instance of this class has the following attributes:

        *   ``file_object`` -- A PLAMS |KFFile| object, referring to the actual RKF file
        *   ``position``    -- The frame to which the cursor is currently pointing in the RKF file
        *   ``mode``        -- Designates whether the file is in read or write mode ('rb' or 'wb')
        *   ``ntap``        -- The number of atoms in the molecular system (needs to be constant throughout)
        *   ``elements``    -- The elements of the atoms in the system (needs to be constant throughout)
        *   ``conect``      -- The connectivity information of the current frame
        *   ``mddata``      -- Read mode only: A dictionary containing data from the MDHistory section in the RKF file
        *   ``read_lattice``-- Read mode only: Wether the lattice vectors will be read from the file
        *   ``read_bonds``  -- Wether the connectivity information will be read from the file
        *   ``saving_freq`` -- How often the 'wb' file is written (default: only when :meth:`close` is called)

        An |RKFTrajectoryFile| object behaves very similar to a regular file object.
        It has read and write methods (:meth:`read_next` and :meth:`write_next`) 
        that read and write from/to the position of the cursor in the ``file_object`` attribute. 
        If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
        the cursor to any frame in the file and reads from there.
        The amount of information stored in memory is kept to a minimum, as only information from the latest frame
        is ever stored.

        Reading and writing to and from the files can be done as follows::

            >>> from scm.plams import RKFTrajectoryFile

            >>> rkf = RKFTrajectoryFile('ams.rkf')
            >>> mol = rkf.get_plamsmol()

            >>> rkfout = RKFTrajectoryFile('new.rkf',mode='wb')

            >>> for i in range(rkf.get_length()) :
            >>>     crd,cell = rkf.read_frame(i,molecule=mol)
            >>>     rkfout.write_next(molecule=mol)
            >>> rkfout.close()

        The above script reads information from the RKF file ``ams.rkf`` into the |Molecule| object ``mol``
        in a step-by-step manner.
        The |Molecule| object is then passed to the :meth:`write_next` method of the new |RKFTrajectoryFile|
        object corresponding to the new rkf file ``new.rkf``.

        The exact same result can also be achieved by iterating over the instance as a callable

            >>> rkf = RKFTrajectoryFile('ams.rkf')
            >>> mol = rkf.get_plamsmol()

            >>> rkfout = RKFTrajectoryFile('new.rkf',mode='wb')

            >>> for crd,cell in rkf(mol) :
            >>>     rkfout.write_next(molecule=mol)
            >>> rkfout.close()

        This procedure requires all coordinate information to be passed to and from the |Molecule| object
        for each frame, which can be time-consuming.
        Some time can be saved by bypassing the |Molecule| object::

            >>> rkf = RKFTrajectoryFile('ams.rkf')

            >>> rkfout = RKFTrajectoryFile('new.rkf',mode='wb')
            >>> rkfout.set_elements(rkf.get_elements())

            >>> for crd,cell in rkf :
            >>>     rkfout.write_next(coords=crd,cell=cell,conect=rkf.conect)
            >>> rkfout.close()

        The only mandatory argument to the :meth:`write_next` method is ``coords``.
        Further time can be saved by setting the ``read_lattice`` and ``read_bonds`` variables to False.

        By default the write mode will create a minimal version of the RKF file, containing only elements,
        coordinates, lattice, and connectivity information.
        This minimal file format can be read by AMSMovie.

        It is possible to store additional information, such as energies, velocities, and charges.
        To enable this, the method :meth:`store_mddata` needs to be called after creation,
        and a dictionary of mddata needs to be passed to the :meth:`write_next` method.
        When that is done, the AMS trajectory analysis tools can be used on the file.
        Restarting an MD run with such a file is however currently not possible::

            >>> rkf = RKFTrajectoryFile('ams.rkf')
            >>> rkf.store_mddata()
            >>> mol = rkf.get_plamsmol()

            >>> rkf_out = RKFTrajectoryFile('new.rkf',mode='wb')
            >>> rkf_out.store_mddata(rkf)

            >>> for i in range(len(rkf)) :
            >>>         crd,cell = rkf.read_frame(i,molecule=mol)
            >>>         rkf_out.write_next(molecule=mol,mddata=rkf.mddata)
            >>> rkf_out.close()
        """
        def __init__ (self, filename, mode='rb', fileobject=None, ntap=None) :
                """
                Initiates an RKFTrajectoryFile object

                * ``filename``   -- The path to the RKF file
                * ``mode``       -- The mode in which to open the RKF file ('rb' or 'wb')
                * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
                * ``ntap``       -- If the file is in write mode, the number of atoms needs to be passed here
                """
                #TODO: If the mddata option is set to True, then the file created here works with AMSMovie and the analysis tools.
                #      To also make is work for restarts, two things have to be added:
                #      1. The final velocities have to be converted from bohr/fs to bohr/au (1/41.341373336493) 
                #         and stored in MDResuts%EndVelocities
                #      2. The final coordinates need to be copied to the Molecule section.

                self.position = 0
                if filename is not None :
                        #fileobject = KFFile(filename,autosave=False,keep_file_open=True)
                        fileobject = KFFile(filename,autosave=False)
                        #fileobject = KFFile(filename,autosave=False,fastsave=True)
                        # This fastsave option (no copying) was not worth it, so I removed it.
                        if fileobject is None :
                                raise PlamsError('KFFile %s not found.'%(rkfname))
                self.file_object = fileobject
                self.mode = mode

                self.ntap = 0
                if ntap is not None :
                        self.ntap = ntap
                self.firsttime = True
                self.coords = numpy.zeros((self.ntap,3))                # Only for reading purposes,
                                                                        # to avoid creating the array each time
                # PLAMS molecule related settings
                self.elements = ['H']*self.ntap
                self.current_molecule = None
                self.store_molecule = True # Even if True, the molecule attribute is only stored during iteration
        
                # RKF specific attributes
                self.nvecs = 3
                self.latticevecs = numpy.zeros((3,3))
                self.read_lattice = True               # Reading time can be saved by skipping the lattice info
                self.read_bonds = True
                self.cell = numpy.zeros((3,3))
                self.conect = None
                self.timestep = None
                self.include_mddata = False
                self.mddata = None
                self.mdunits = None
                self.include_historydata = False       # Any additional data along the history section will be stored
                self.historydata = None
                self.saving_freq = None                # By default the 'wb' file is only written upon closing
                                                       # Saving more often is much slower.
                # Skip to the trajectory part of the file (only if in read mode, because coords are required in header)
                if self.mode == 'rb' :
                        self._read_header()
                elif self.mode == 'wb' :
                        sections = self.file_object.sections()
                        if len(sections) > 0 : 
                                raise PlamsError ('RKF file %s already exists'%(filename))
                else :
                        raise PlamsError ('Mode %s is invalid. Only "rb" and "wb" are allowed.'%(self.mode))

        def store_mddata (self, rkf=None) :
                """
                Read/write an MDHistory section

                * ``rkf`` -- If in write mode an RKFTrajectoryFile object in read mode needs to be passed to extract unit info
                """
                self.include_mddata = True
                if 'r' in self.mode :
                        self._set_mddata_units()
                elif 'w' in self.mode :
                        if rkf is not None :
                                self.timestep = rkf.timestep
                                self._set_mdunits(rkf.mdunits)

        def store_historydata (self) :
                """
                Read/write non-standard entries in the History section
                """
                self.include_historydata = True

        def close (self) :
                """
                Execute all prior commands and cleanly close and garbage collect the RKF file
                """
                # Write the step info
                if self.timestep is not None and self.mode == 'wb' :
                        self.file_object.write('MDResults','StartStep',0)
                        self.file_object.write('MDResults','StartTime[fs]',0.)
                        nsteps = self.get_length()
                        self.file_object.write('MDResults','EndStep',nsteps-1)
                        self.file_object.write('MDResults','EndTime[fs]',(nsteps-1)*self.timestep)

                # Write to file
                if self.mode == 'wb' :
                        # First write the last frame into the Molecule section  
                        self._rewrite_molecule()
                        # Then write to file
                        self.file_object.save()
                del(self)

        def _rewrite_molecule (self) :
                """
                Overwrite the molecule section with the latest frame
                """
                crd,cell = self.read_last_frame()
                self._write_molecule_section(crd,cell)

        def _read_header (self, molecule_section='Molecule') :
                """
                Set up info required for reading frames
                """
                self.elements = self.file_object.read(molecule_section,'AtomSymbols')
                # If read from memory and not from file (write mode), it is already a list
                if isinstance(self.elements,str) :
                        self.elements = self.elements.split()
                self.elements = [el.split('.')[0] for el in self.elements]
                if self.file_object.reader is not None :
                        if ('MDHistory','Time(1)') in self.file_object :
                                times = self.file_object.read('MDHistory','Time(1)')
                                if isinstance(times,list) :
                                        self.timestep = times[1]
                self.ntap = len(self.elements)
                self.coords = numpy.zeros((self.ntap,3))
                try :
                        self.latticevecs = numpy.array(self.file_object.read(molecule_section,'LatticeVectors'))
                        #self.nvecs = int(len(self.cell)/3)
                        # New code 27-05-2020
                        self.nvecs = int(len(self.latticevecs)/3) # Why did I remove this line locally?!
                        self.latticevecs = self.latticevecs.reshape((self.nvecs,3))
                except (KeyError,AttributeError) :
                        pass

        def _set_mddata_units (self) :
                """
                Get the units for the mddata, if those are to be read
                """
                # Look for the items
                section = 'MDHistory'
                sections = self.file_object.get_skeleton()
                item_keys = [kn for kn in sections[section] if 'ItemName' in kn]
                items = [self.file_object.read(section,kn) for kn in item_keys]

                # Get the data for each item
                unit_dic = {}
                for item in items :
                        if '%s(units)'%(item) in self.file_object.reader._sections[section] :
                                unit_dic[item] = self.file_object.read(section,'%s(units)'%(item))

                self.mdunits = unit_dic

        def _write_header (self, coords, cell, molecule=None) :
                """
                Write Molecule info to file (elements, periodicity)
                """
                # First write the general section
                self._write_general_section()

                # Then write the input molecule
                self._update_celldata(cell)
                self._write_molecule_section(coords, cell, molecule=molecule)
                self._write_molecule_section(coords, cell, section='InputMolecule', molecule=molecule)
                if self.include_mddata :
                        # Start setting up the MDHistory section as well
                        self.file_object.write('MDHistory','blockSize',100)

                # Now make sure that it is possible to read from the file as well
                self._read_header()

        def _write_general_section (self) :
                """
                Write the General section of the RKF file
                """ 
                self.file_object.write('General','file-ident','RKF')
                self.file_object.write('General','termination status','NORMAL TERMINATION')
                self.file_object.write('General','program','ams')
                self.file_object.write('General','user input',' ')

        def _update_celldata (self, cell) :
                """
                Use the newly supplied cell to update the dimensionality of the system
                """
                shape = numpy.array(cell).shape
                if len(shape) == 2 :
                        self.nvecs = shape[0]
                        self.latticevecs = numpy.zeros((self.nvecs,3)) # Not really necessay

        def _write_molecule_section (self, coords, cell, section='Molecule', molecule=None) :
                """
                Write the molecule section
                """
                # Then write the input molecule
                charge = 0.
                element_numbers = [PeriodicTable.get_atomic_number(el) for el in self.elements]

                self.file_object.write(section,'nAtoms',len(self.elements))
                self.file_object.write(section,'AtomicNumbers',element_numbers)
                self.file_object.write(section,'AtomSymbols',self.elements)
                crd = [Units.convert(float(c),'angstrom','bohr') for coord in coords for c in coord]
                self.file_object.write(section,'Coords',crd)
                self.file_object.write(section,'Charge',charge)
                if cell is not None :
                        self.file_object.write(section,'nLatticeVectors',self.nvecs)
                        vecs = [Units.convert(float(v),'angstrom','bohr') for vec in cell for v in vec]
                        self.file_object.write(section,'LatticeVectors',vecs)
                # Should it write bonds?
                # Write atom properties
                if molecule is not None :
                        suffixes = []
                        present = False
                        for at in molecule :
                                if 'suffix' in at.properties :
                                        present = True
                                        suffixes.append(at.properties.suffix)
                                else :
                                        suffixes.append('')
                        if present :
                                self.file_object.write(section,'EngineAtomicInfo','\x00'.join(suffixes))

        def _set_mdunits (self, mdunits) :
                """
                Store the dictionary with MD Units
                """
                if self.include_mddata :
                        self.mdunits = mdunits

        def get_plamsmol (self) :
                """
                Extracts a PLAMS molecule object from the RKF file
                """
                section_dict = self.file_object.read_section('InputMolecule')
                if len(section_dict) == 0 :
                        section_dict = self.file_object.read_section('Molecule')
                plamsmol = Molecule._mol_from_rkf_section(section_dict)
                return plamsmol

        def read_frame (self, i, molecule=None) :
                """
                Reads the relevant info from frame ``i`` and returns it, or stores it in ``molecule``

                * ``i``        -- The frame number to be read from the RKF file
                * ``molecule`` -- |Molecule| object in which the new coordinates need to be stored
                """
                # Read the cell data
                cell = None
                if self.read_lattice :
                        try :
                                cell = self._read_cell_data (i)
                        except (KeyError,AttributeError) :
                                pass

                # Read the bond data
                conect = None
                bonds = None
                if self.read_bonds :
                        conect, bonds = self._read_bond_data(section='History', step=i)
                self.conect = conect

                # Read the coordinates, and possible pass them to molecule
                try :
                        self._read_coordinates(i, molecule, cell, bonds)
                        # This has changed self.coords behind the scenes
                except (KeyError,AttributeError) :
                        return None, None

                # Read and store any additional data in the history section
                if self.include_historydata :
                        self._store_historydata_for_step(i)
                # Read and store all MDData for this frame
                if self.include_mddata :
                        self._store_mddata_for_step(i)
                # Finalize
                if self.firsttime :
                        self.firsttime = False

                self.position = i
                return self.coords, cell

        def _read_coordinates (self, i, molecule, cell, bonds) :
                """
                Read the coordinates at step i, and possible pass them to molecule
                """
                if not self.coords.shape == (self.ntap,3) :
                        raise PlamsError('coords attribute has been changed outside the class')
                coords = self.coords.reshape(self.ntap*3)
                coords[:] = self.file_object.read('History', 'Coords(%i)'%(i+1))
                coords *= bohr_to_angstrom
                # This has changed self.coords behind the scenes

                # Create the molecule
                if isinstance(molecule,Molecule) :
                        cell_reduced = None
                        if cell is not None :
                                cell_reduced = cell[:self.nvecs]
                        self._set_plamsmol(self.coords,cell_reduced,molecule,bonds)

        def _read_cell_data (self, i) :
                """
                Read the cell data at step i
                """
                latticevecs = self.latticevecs.reshape(self.nvecs*3)
                latticevecs[:] = self.file_object.read('History','LatticeVectors(%i)'%(i+1)) #* bohr_to_angstrom
                latticevecs *= bohr_to_angstrom
                # This changed self.latticevecs behind the scenes
                #self.cell[:self.nvecs] = latticevecs
                self.cell[:self.nvecs] = self.latticevecs
                cell = self.cell
                return cell

        def _read_bond_data (self, section, step=None) :
                """
                Read the bond data from the rkf file
                """
                conect = None
                bonds = None
                try :
                        step_txt = ''
                        if step is not None :
                                step_txt = '(%i)'%(step+1)
                        #if ('History','Bonds.Index%s'%(step_txt)) in self.file_object :
                        indices = self.file_object.read(section,'Bonds.Index%s'%(step_txt))
                        connection_table = self.file_object.read(section,'Bonds.Atoms%s'%(step_txt))
                        if isinstance(connection_table,int) :
                                connection_table = [connection_table]
                        bond_orders = self.file_object.read(section,'Bonds.Orders%s'%(step_txt))
                        if isinstance(bond_orders,float) :
                                bond_orders = [bond_orders]
                        conect = {}
                        bonds = []
                        for i,(start,end) in enumerate(zip(indices[:-1],indices[1:])) :
                                if end-start > 0 :
                                        conect[i+1] = connection_table[start-1:end-1]
                                        for j in range(start-1,end-1) :
                                                bonds.append([i+1,connection_table[j],bond_orders[j]])
                        # Now correct the connection table
                        conect_sym = {}
                        for i, neighbors_i in conect.items() :
                                conect_sym[i] = neighbors_i
                                for j in neighbors_i :
                                        if not j in conect_sym.keys() :
                                                conect_sym[j] = []
                                        if j in conect.keys() : 
                                                conect_sym[j] = conect[j]
                                        if not i in conect_sym[j] :
                                                conect_sym[j].append(i)
                        conect = conect_sym     
                except (KeyError, AttributeError) :
                        pass
                return conect, bonds

        def _store_mddata_for_step (self, istep) :
                """
                Store the data from the MDHistory section
                """
                if self.mddata == None : self.mddata = {}
                section = 'MDHistory'

                # First get the block info
                blocksize = self.file_object.read(section, 'blockSize')
                nblocks = self.file_object.read(section, 'nBlocks')

                # Look for the items
                sections = self.file_object.get_skeleton()
                item_keys = [kn for kn in sections[section] if 'ItemName' in kn]
                items = [self.file_object.read(section,kn) for kn in item_keys]

                # Get the data for each item
                for item in items :
                        # First read the units (not needed?)
                        #units = ''
                        #if '%s(units)'%(item) in self.file_object.reader._sections[section] :
                        #        units = self.file_object.read(section,'%s(units)'%(item))

                        dim = self.file_object.read(section,'%s(dim)'%(item))
                        if dim == 1 and not self.file_object.read(section,'%s(perAtom)'%(item)):
                                # Stored in block format
                                block = int(istep/blocksize)
                                pos = istep%blocksize
                                values = self.file_object.read(section,'%s(%i)'%(item,block+1))
                                if not isinstance(values,list) : values = [values]
                                self.mddata[item] = values[pos]
                        else :  
                                self.mddata[item] = self.file_object.read(section,'%s(%i)'%(item,istep+1))

        def _store_historydata_for_step (self, istep) :
                """
                Store the extra data from the History section
                """
                if self.historydata == None : self.historydata = {}
                section = 'History'

                # Look for the items
                sections = self.file_object.get_skeleton()
                item_keys = [kn for kn in sections[section] if 'ItemName' in kn]
                items = [self.file_object.read(section,kn) for kn in item_keys]
                standard_items = ['Coords','nLatticeVectors','LatticeVectors','Bonds.Index','Bonds.Atoms','Bonds.Orders']
                items = [item for item in items if not item in standard_items]

                # Get the data for each item
                for item in items :
                        # Never stored in block format
                        self.historydata[item] = self.file_object.read(section,'%s(%i)'%(item,istep+1))

        def _is_endoffile (self) :
                """
                Reads and checks If the end of file is reached.
                """
                return ('History', 'Coords(%i)'%(self.position+1)) in self.file_object

        def read_next (self, molecule=None, read=True) :
                """
                Reads coordinates and lattice vectors from the current position of the cursor and returns it

                * ``molecule`` -- |Molecule| object in which the new coordinates need to be stored
                * ``read``     -- If set to False the cursor will move to the next frame without reading
                """
                if not read and not self.firsttime :
                        return self._move_cursor_without_reading()

                if self.firsttime :
                        self.firsttime = False

                crd, vecs = self.read_frame (self.position,molecule)
                self.position += 1
                return crd, vecs 

        def write_next (self, coords=None, molecule=None, cell=[0.,0.,0.], conect=None, historydata=None, mddata=None) :
                """
                Write frame to next position in trajectory file

                * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates in angstrom
                * ``molecule`` -- A molecule object to read the molecular data from
                * ``cell``     -- A set of lattice vectors (or cell diameters for an orthorhombic system) in angstrom
                * ``conect``   -- A dictionary containing the connectivity info (e.g. {1:[2],2:[1]})
                * ``conect``   -- A dictionary containing the connectivity info (e.g. {1:[2],2:[1]})
                * ``historydata`` -- A dictionary containing additional variables to be written to the History section
                * ``mddata``   -- A dictionary containing the variables to be written to the MDHistory section

                The ``mddata`` dictionary can contain the following keys:
                ('TotalEnergy', 'PotentialEnergy', 'Step', 'Velocities', 'KineticEnergy', 
                'Charges', 'ConservedEnergy', 'Time', 'Temperature')

                The ``historydata`` dictionary can contain for example:
                ('Energy','Gradients','StressTensor')
                All values must be in atomic units
                Numpy arrays or lists of lists will be flattened before they are written to the file

                .. note::

                        Either ``coords`` or ``molecule`` are mandatory arguments
                """
                # Check for common error in the arguments
                if coords is not None :
                        if isinstance(coords,Molecule) :
                                raise PlamsError('The PLAMS molecule needs to be passed as the second argument (molecule)')

                if isinstance(molecule,Molecule) :
                        coords, cell, elements, conect, props = self._read_plamsmol(molecule)
                        if self.position == 0 : self.elements = elements
                # Make sure that the cell consists of vectors
                cell = self._convert_cell(cell)
                if conect is not None :
                        if len(conect) == 0 : conect = None
                self.conect = conect

                # Include a check on the size of coords?
                if len(coords) != len(self.elements) :
                        raise PlamsError('The coordinates do not match the rest of the trajectory')

                # If this is the first step, write the header
                if self.position == 0 :
                        self._write_header(coords,cell,molecule)

                # Define some local variables
                step = self.position
                if mddata is not None :
                        if 'Step' in mddata :
                                step = mddata['Step']
                # Energy should be read from mddata first, otherwise from historydata, otherwise set to zero
                energy = self._set_energy(mddata, historydata)
                if not self.include_historydata :
                        historydata = {}
                historydata['Energy'] = energy

                # Write the history section
                counter = 1
                counter = self._write_history_entry(step, coords, cell, conect, historydata, counter)

                if self.include_mddata and mddata is not None :
                        self._write_mdhistory_entry(mddata)

                self.position += 1

                if self.saving_freq is not None :
                        if self.position%self.saving_freq == 0 : self.file_object.save()

        def _set_energy (self, mddata, historydata) :
                """
                Looks if an energy is passed as input, and it not, sets to zero
                """
                energy = None
                if mddata is not None :
                        if 'PotentialEnergy' in mddata :
                                energy = mddata['PotentialEnergy']
                if energy is None :
                        if historydata is not None :
                                if 'Energy' in historydata :
                                        energy = historydata['Energy']
                if energy is None :     
                        energy = 0.
                return energy

        def _write_history_entry (self, step, coords, cell, conect, historydata=None, counter=1) :
                """
                Write the full entry into the History section
                """
                self.file_object.write('History','nEntries',self.position+1)
                self.file_object.write('History','currentEntryOpen',False)
                self._write_keydata_in_history('Step', counter, False, 1, self.position+1, step)
                counter += 1
                crd = [float(c)/bohr_to_angstrom for coord in coords for c in coord]
                self._write_keydata_in_history('Coords', counter, True, 3, self.position+1, crd)
                counter += 1
                #self._write_keydata_in_history('Energy', counter, False, 1, self.position+1, energy)
                #counter += 1
                if cell is not None :
                        self._write_keydata_in_history('nLatticeVectors', counter, False, 1, self.position+1, self.nvecs)
                        counter += 1
                        vecs = [float(v)/bohr_to_angstrom for vec in cell for v in vec]
                        # I should probably rethink the dimension of the lattice vectors (generalize it)
                        self._write_keydata_in_history('LatticeVectors', counter, False, [3,3], self.position+1, vecs)
                        counter += 1

                if historydata is not None :
                         counter = self._write_dictionary_to_history(historydata,'History',counter)
                #if gradients is not None :
                #        grd = [float(g) for grad in gradients for g in grad]
                #        self._write_keydata_in_history('Gradients', counter, True, 3, self.position+1, grd)

                #if stresstensor is not None :
                #        stre = [float(s) for stress in stresstensor for s in stress]
                #        self._write_keydata_in_history('StressTensor', counter, False, [3,3], self.position+1, stre)

                # Write the bond info
                if conect is not None :
                        counter = self._write_bonds_in_history(conect, counter, len(coords))

                return counter

        def _write_bonds_in_history (self, conect, counter, nats) :
                """
                Write the bond data into the history section
                """
                # Create the index list (correct for double counting)
                connection_table = [[at for at in conect[i+1] if at>i+1] if i+1 in conect else [] for i in range(nats)]
                numbonds = [len(neighbors) for neighbors in connection_table]

                indices = [sum(numbonds[:i])+1 for i in range(nats+1)]
                self.file_object.write('History','Bonds.Index(%i)'%(self.position+1),indices)
                self.file_object.write('History','ItemName(%i)'%(counter),'%s'%('Bonds.Index'))
                counter += 1

                # Flatten the connection table
                connection_table = [at for i in range(nats) for at in connection_table[i]]
                self.file_object.write('History','Bonds.Atoms(%i)'%(self.position+1),connection_table)
                self.file_object.write('History','ItemName(%i)'%(counter),'%s'%('Bonds.Atoms'))
                counter += 1

                bond_orders = [1.0 for bond in connection_table]
                self.file_object.write('History','Bonds.Orders(%i)'%(self.position+1),bond_orders)
                self.file_object.write('History','ItemName(%i)'%(counter),'%s'%('Bonds.Orders'))
                counter += 1

                return counter

        def _write_mdhistory_entry (self, mddata) :
                """
                Write the entry in the MDHistory section
                """
                counter = 1
                counter = self._write_dictionary_to_history(mddata,'MDHistory',counter)

        def _write_dictionary_to_history (self, data, section, counter=1) :
                """
                Add the entries of a dictionary to a History section 
                """
                self.file_object.write(section,'nEntries',self.position+1)
                self.file_object.write(section,'currentEntryOpen',False)
                for key, var in data.items() :
                        # Make sure that the entry is either a scalar or a 1D list
                        var = self._flatten_variable(var)
                        peratom = False
                        dim = 1
                        if isinstance(var,list) :
                                if len(var) % len(self.elements) == 0 :
                                        dim = int(len(var) / len(self.elements))
                                        peratom = True
                                else :
                                        dim = len(var)
                        self._write_keydata_in_history(key, counter, peratom, dim, self.position+1, var, section)
                        counter += 1
                return counter

        def _flatten_variable (self, var) :
                """
                Make sure that the variable is a Python 1D list (not numpy)
                """
                while(1) :
                        if isinstance(var,list) or isinstance(var,numpy.ndarray) :
                                if len(var) == 0 : break
                                if isinstance(var[0],list) or isinstance(var[0],numpy.ndarray) :
                                        var = [v for varitem in var for v in varitem]
                                else :
                                        if isinstance(var[0],numpy.int64) :
                                                var = [int(v) for v in var]
                                        elif isinstance(var[0],numpy.float64) :
                                                var = [float(v) for v in var]
                                        break
                        else :
                                if isinstance(var,numpy.int64) : 
                                        var = int(var)
                                elif isinstance(var,numpy.float64) :
                                        var = float(var)
                                break
                return var

        def _write_keydata_in_history(self, key, i, perAtom, dim, step, values, section='History') :
                """
                Write all data about a key value in KFFile
                """
                # Some data only needs to be printed once
                printstartdata = False
                if step == 1 :
                        printstartdata = True

                # Block code: if the data is to be written as blocks, then step and values need to be replaced.
                if section == 'MDHistory' :
                        step, values = self._get_block_info (key, perAtom, dim, step, values, section)
                                
                # The rest should be independent on format (block or individual)
                self.file_object.write(section,'%s(%i)'%(key,step),values)
                if printstartdata :
                        self.file_object.write(section,'ItemName(%i)'%(i),'%s'%(key))
                        self.file_object.write(section,'%s(perAtom)'%(key),perAtom)
                        self.file_object.write(section,'%s(dim)'%(key),dim)
                        if section == 'MDHistory' and self.mdunits is not None :
                                if key in self.mdunits :
                                        self.file_object.write(section,'%s(units)'%(key),self.mdunits[key])

        def _get_block_info (self, key, perAtom, dim, step, values, section) :
                """
                If the data is to be written as blocks, then step and values need to be replaced.
                """
                if dim==1 and not perAtom :
                        try :
                                blocksize = self.file_object.read(section,'blockSize')
                        except AttributeError :
                                raise Exception ('Set include_mddata to write the MD section.')
                        iblock = int((step-1)/blocksize) + 1
                        if step%blocksize != 1 :
                                try :
                                        old_values = self.file_object.read(section,'%s(%i)'%(key,iblock))
                                        if not isinstance(old_values,list) :
                                                old_values = [old_values]
                                except AttributeError :
                                        old_values = []
                                values = old_values + [values] # Values is a scalar
                        else :
                                self.file_object.write(section,'nBlocks',iblock)
                        step = iblock
                return step, values

        def rewind (self, nframes=None) :
                """
                Rewind the file either by ``nframes`` or to the first frame

                *   ``nframes`` -- The number of frames to rewind
                """
                self.firsttime = True
                self.position = 0

        def get_length (self) :
                """
                Get the number of frames in the file
                """
                nsteps = self.file_object.read('History', 'nEntries')
                return nsteps

        def read_last_frame (self, molecule=None) :
                """
                Reads the last frame from the file
                """
                nsteps = self.get_length()
                crd,cell = self.read_frame(nsteps-1, molecule)
                return crd, cell



#!/usr/bin/env python

import numpy
from ..mol.molecule import Molecule
from ..tools.geometry import cell_shape
from ..tools.geometry import cellvectors_from_shape
from .trajectoryfile import TrajectoryFile

__all__ = ['XYZTrajectoryFile','create_xyz_string']

class XYZTrajectoryFile (TrajectoryFile) :
        """
        Class representing an XYZ file containing a molecular trajectory

        An instance of this class has the following attributes:

        *   ``file_object`` -- A Python :py:class:`file` object, referring to the actual XYZ file
        *   ``position``    -- The frame to which the cursor is currently pointing in the XYZ file
        *   ``mode``        -- Designates whether the file is in read or write mode ('r' or 'w')
        *   ``ntap``        -- The number of atoms in the molecular system (needs to be constant throughout)
        *   ``elements``    -- The elements of the atoms in the system (needs to be constant throughout)

        An |XYZTrajectoryFile| object behaves very similar to a regular file object.
        It has read and write methods (:meth:`read_next` and :meth:`write_next`) 
        that read and write from/to the position of the cursor in the ``file_object`` attribute. 
        If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
        the cursor to any frame in the file and reads from there.
        The amount of information stored in memory is kept to a minimum, as only information from the current frame
        is ever stored.

        Reading and writing to and from the files can be done as follows::

            >>> from scm.plams import XYZTrajectoryFile

            >>> xyz = XYZTrajectoryFile('old.xyz')
            >>> mol = xyz.get_plamsmol()

            >>> xyzout = XYZTrajectoryFile('new.xyz',mode='w')

            >>> for i in range(xyz.get_length()) :
            >>>     crd,cell = xyz.read_frame(i,molecule=mol)
            >>>     xyzout.write_next(molecule=mol)

        The above script reads information from the XYZ file ``old.xyz`` into the |Molecule| object ``mol``
        in a step-by-step manner.
        The |Molecule| object is then passed to the :meth:`write_next` method of the new |XYZTrajectoryFile|
        object corresponding to the new xyz file ``new.xyz``.

        The exact same result can also be achieved by iterating over the instance as a callable

            >>> xyz = XYZTrajectoryFile('old.xyz')
            >>> mol = xyz.get_plamsmol()

            >>> xyzout = XYZTrajectoryFile('new.xyz',mode='w')

            >>> for crd,cell in xyz(mol) :
            >>>     xyzout.write_next(molecule=mol)

        This procedure requires all coordinate information to be passed to and from the |Molecule| object
        for each frame, which can be time-consuming.
        It is therefore also possible to bypass the |Molecule| object when reading through the frames::

            >>> xyz = XYZTrajectoryFile('old.xyz')

            >>> xyzout = XYZTrajectoryFile('new.xyz',mode='w')
            >>> xyzout.set_elements(xyz.get_elements())

            >>> for crd,cell in xyz :
            >>>     xyzout.write_next(coords=crd)
            >>> xyzout.close()

        By default the write mode will create a minimal version of the XYZ file, containing only elements
        and coordinates. 
        Additional information can be written to the file by supplying additional arguments
        to the :meth:`write_next` method. 
        The additional keywords `step` and `energy` trigger the writing of a remark containing
        the molecule name, the step number, the energy, and the lattice vectors.

            >>> mol = Molecule('singleframe.xyz')

            >>> xyzout = XYZTrajectoryFile('new.xyz',mode='w')
            >>> xyzout.set_name('MyMol')

            >>> xyzout.write_next(molecule=mol, step=0, energy=5.)
        """
        def __init__ (self, filename, mode='r', fileobject=None, ntap=None) :
                """
                Initiates an XYZTrajectoryFile object

                * ``filename``   -- The path to the XYZ file
                * ``mode``       -- The mode in which to open the XYZ file ('r' or 'w')
                * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
                * ``ntap``       -- If the file is in write mode, the number of atoms needs to be passed here
                """
                TrajectoryFile.__init__(self, filename, mode, fileobject, ntap)

                # XYZ specific attributes
                self.name = 'PlamsMol'

                # Required setup before frames can be read/written
                if self.mode == 'r' :
                        self._read_header()

                # Specific XYZ stuff
                self.include_historydata = False
                self.historydata = None

        def store_historydata (self) :
                """
                Additional data should be read from/written to file
                """
                self.include_historydata = True

        def set_name (self, name) :
                """
                Sets the name of the system, in case an extensive write is requested

                *   ``name`` -- A string containing the name of the molecule
                """
                self.name = name

        def _read_header (self) :
                """
                Set up info required for reading frames
                """
                line = self.file_object.readline()
                self.ntap = int(line.split()[0])
                if self.coords.shape == (0,3) :
                        self.coords = numpy.zeros((self.ntap,3)) 

                self.file_object.readline()

                elements = []
                for i in range(self.ntap) :
                        line = self.file_object.readline()
                        elements.append(line.split()[0])
                self.elements = elements

                self.file_object.seek(0)

        def read_next (self, molecule=None, read=True) :
                """
                Reads coordinates from the current position of the cursor and returns it

                * ``molecule`` -- |Molecule| object in which the new coordinates need to be stored
                * ``read``     -- If set to False the cursor will move to the next frame without reading
                """
                if not read and not self.firsttime :
                        return self._move_cursor_without_reading()

                cell = None
                # Read the coordinates
                crd, cell = self._read_coordinates(molecule)
                if crd is None :
                        return None, None       # End of file is reached

                if self.firsttime :
                        self.firsttime = False

                self.position += 1
                
                return self.coords, cell

        def _read_coordinates (self, molecule) :
                """
                Read the coordinates from file, and place them in the molecule
                """
                cell = None
                for i in range(2) :
                        line = self.file_object.readline()
                        if len(line) == 0 :
                                return None, None           # End of file is reached
                # Handle the comment line
                historydata = data_from_xyzcomment(line)
                if 'Lattice' in historydata :
                        cell = historydata['Lattice']
                        del historydata['Lattice']
                if self.include_historydata :
                        self.historydata = historydata

                # Write coordinates
                for i in range(self.ntap) :
                        line = self.file_object.readline()
                        self.coords[i,:] = [float(w) for w in line.split()[1:4]]

                if isinstance(molecule,Molecule) :
                        self._set_plamsmol(self.coords, cell, molecule)

                return self.coords, cell

        def _is_endoffile (self) :
                """
                If the end of file is reached, return coords and cell as None
                """
                end = False
                for i in range(self.ntap+2) :
                        line = self.file_object.readline()
                        if len(line) == 0 :
                                end = True
                                break
                return end

        def write_next (self,coords=None,molecule=None,cell=[0.,0.,0.],conect=None,historydata=None) :
                """
                Write frame to next position in trajectory file

                * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates
                * ``molecule`` -- A molecule object to read the molecular data from
                * ``cell``     -- A set of lattice vectors or cell diameters
                * ``conect``   -- A dictionary containing connectivity info (not used)
                * ``historydata`` -- A dictionary containing additional variables to be written to the comment line

                The ``historydata`` dictionary can contain for example:
                ('Step','Energy'), the frame number and the energy respectively

                .. note::

                        Either ``coords`` or ``molecule`` are mandatory arguments
                """
                if isinstance(molecule,Molecule) :
                        coords, cell, elements = self._read_plamsmol(molecule)[:3]
                        if self.position == 0 :
                                self.elements = elements
                cell = self._convert_cell(cell)

                self._write_moldata(coords, cell, historydata)

                self.position += 1

        def _write_moldata (self, coords, cell, historydata) :
                """
                Write all molecular info to file
                """
                if historydata is None :
                        historydata = {}
                if self.include_historydata and len(historydata) > 0 :
                        step = self.position
                        if 'Step' in historydata :
                                step = historydata['Step']
                        energy = 0.
                        if 'Energy' in historydata :
                                energy = historydata['Energy']
                        box = None
                        if cell is not None :
                                #box = PDBMolecule().box_from_vectors(cell)
                                box = cell_shape(cell)
                        name = self.name
                        if 'Name' in historydata :
                                name = historydata['Name']
                        block = create_xyz_string(self.elements,coords,energy,box,step,name)
                else :
                        block = create_xyz_string(self.elements, coords)
                self.file_object.write(block)

        def _rewind_to_first_frame(self) :
                """
                Rewind the file to the first frame
                """
                self.file_object.seek(0)
                self.firsttime = True
                self.position = 0

        def _rewind_n_frames(self,nframes) :
                """
                Rewind the file by nframes frames
                """
                new_frame = self.position - nframes
                self._rewind_to_first_frame()
                for i in range(new_frame) :
                        self.read_next(read=False)


def create_xyz_string (elements, coords, energy=None, box=None, step=None, name='PlamsMol') :
        """
        Write an XYZ file based on the elements and the coordinates of the atoms
        """
        block = '%8i\n'%(len(elements))
        if step is not None :
                if energy is None : energy = 0.
                comment = '%-40s%6i %16.6f'%(name, step, energy)
                if box is not None :
                        for value in box :
                                comment += '%7.2f'%(value)
                block += comment
        block += '\n'
        for el,crd in zip(elements,coords) :
                block += '%8s '%(el)
                for x in crd :
                        block += '%20.10f '%(x)
                block += '\n'
        return block

def data_from_xyzcomment (line) :
        """
        Convert XYZ comment line (cell angles are in radians)
        """
        words = line.split()
        if len(words) == 0 : return {}
        if len(words) < 3 : return {'Line':line}
        xyzdic = {}
        xyzdic['Name'] = words[0]
        try :
                xyzdic['Step'] = int(words[1])
                xyzdic['Energy'] = float(words[2])
        except ValueError :
                xyzdic['Line'] = line
                return xyzdic
        if len(words) >= 6 :
                try :
                        box = [float(w) for w in words[3:6]]
                except ValueError :
                        return xyzdic
                if len(words) >= 9 :
                        try :
                                angles = [float(w) for w in words[6:9]]
                        except ValueError :
                                return xyzdic
                        angles = [a for a in angles]
                        box += angles
                lattice = cellvectors_from_shape(box)
                xyzdic['Lattice'] = lattice
        return xyzdic



#!/usr/bin/env python

import os
import struct
import array
import numpy
from ..mol.molecule import Molecule
from ..core.errors import PlamsError
from .trajectoryfile import TrajectoryFile

__all__ = ['DCDTrajectoryFile']

class DCDTrajectoryFile (TrajectoryFile) :
        """
        Class representing a DCD file containing a molecular trajectory

        An instance of this class has the following attributes:

        *   ``file_object`` -- A Python :py:class:`file` object, referring to the actual DCD file
        *   ``position``    -- The frame to which the cursor is currently pointing in the DCD file
        *   ``mode``        -- Designates whether the file is in read or write mode ('rb' or 'wb')
        *   ``ntap``        -- The number of atoms in the molecular system (needs to be constant throughout)

        An |DCDTrajectoryFile| object behaves very similar to a regular file object.
        It has read and write methods (:meth:`read_next` and :meth:`write_next`) 
        that read and write from/to the position of the cursor in the ``file_object`` attribute. 
        If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
        the cursor to any frame in the file and reads from there.
        The amount of information stored in memory is kept to a minimum, as only information from the current frame
        is ever stored.

        Reading and writing to and from the files can be done as follows::

            >>> from scm.plams import DCDTrajectoryFile

            >>> dcd = DCDTrajectoryFile('old.dcd')
            >>> mol = dcd.get_plamsmol()

            >>> dcdout = DCDTrajectoryFile('new.dcd',mode='wb',ntap=dcd.ntap)

            >>> for i in range(dcd.get_length()) :
            >>>     crd,cell = dcd.read_frame(i,molecule=mol)
            >>>     dcdout.write_next(molecule=mol)

        The above script reads information from the DCD file old.dcd into the |Molecule| object ``mol``
        in a step-by-step manner.
        The |Molecule| object is then passed to the :meth:`write_next` method of the new |DCDTrajectoryFile|
        object corresponding to the new dcd file ``new.dcd``.

        The exact same result can also be achieved by iterating over the instance as a callable

            >>> dcd = DCDTrajectoryFile('old.dcd')
            >>> mol = dcd.get_plamsmol()

            >>> dcdout = DCDTrajectoryFile('new.dcd',mode='wb',ntap=dcd.ntap)

            >>> for crd,cell in dcd(mol) :
            >>>     dcdout.write_next(molecule=mol)

        This procedure requires all coordinate information to be passed to and from the |Molecule| object
        for each frame, which can be time-consuming.
        It is therefore also possible to bypass the |Molecule| object when reading through the frames::

            >>> dcd = DCDTrajectoryFile('old.dcd')

            >>> dcdout = DCDTrajectoryFile('new.dcd',mode='wb',ntap=dcd.ntap)

            >>> for crd,cell in dcd :
            >>>     dcdout.write_next(coords=crd,cell=cell)
        """

        def __init__ (self, filename=None, mode='rb', fileobject=None, ntap=None) :
                """
                Initiates a DCDTrajectoryFile object

                * ``filename``   -- The path to the DCD file
                * ``mode``       -- The mode in which to open the DCD file ('rb' or 'wb')
                * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
                * ``ntap``       -- If the file is in write mode, the number of atoms needs to be passed here
                """
                TrajectoryFile.__init__(self, filename, mode, fileobject, ntap)

                # DCD specific attributes
                self.celldata = False                   # Whether the cell data is in the file (only set in read-mode)
                self.byte_order = '@'                   # native endian
                self.timesteps = (0,0,0)                # (nsteps,startstep,steps_between_saves)
                self.delta = 2.0                        # The timestep (if stored)
                self.namnf = 0                          # The number of free atoms (if stored)
                self.stepsize = 0                       # The number of bytes used to store a single conformation
                self.intsize = struct.calcsize('i')     # The number of bytes in an integer
                self.floatsize = struct.calcsize('f')   # The number of bytes in a float

                # Skip to the trajectory part of the file
                if self.mode == 'rb' :
                        self._read_header()
                #elif self.mode == 'wb' :
                #        self._write_header()

        def set_byteorder (self, byteorder) :
                """
                Specifies wether the file to be read from or written to is little endian or big endian

                * ``byteorder`` -- Can be either '@', '<', or '>', denoting native endian, little endian, or big endian
                """
                self.byte_order = byteorder

        def _read_variable (self,size) :
                """
                Reads a variable from a binary file

                :note:: 
                        Assumes that the variable has been written on a machine
                        with the same bitsizes for the data
                """
                if size == 'i' :
                        varsize = self.intsize
                if size == 'f' :
                        varsize = self.floatsize
                else :
                        varsize = struct.calcsize(size)
                var_char = self.file_object.read(varsize)
                fmt = self.byte_order+size
                var = struct.unpack(fmt,var_char)[0]

                return var

        def _write_variable (self,var,size) :
                """
                Writes a variable to a binary file
                """
                var_char = struct.pack(size,var)
                self.file_object.write(var_char)

        def _read_header (self) :
                """
                Reads in the header information in the binary DCD file
                """
                # This should be an int32
                input_integer = self._read_variable('i')
                if input_integer != 84 :
                        print(input_integer)
                        print('BAD DCD FORMAT')
                        raise PlamsError

                # Read CORD string
                car4 = self.file_object.read(4).decode()
                if car4 != 'CORD' :
                        print('BAD DCD FORMAT')
                        raise PlamsError

                # Here we read the data of the timesteps
                # NSET: The number of sets of coordinates
                # ISTART: The starting timestep
                # NSAVC: The number of timesteps between dcd saves
                timesteps = array.array('i')
                timesteps.fromfile(self.file_object,3)
                self.timesteps = tuple(timesteps)
                #print 'timesteps: ',self.timesteps

                # 5 blanc integers
                dummies = array.array('i')
                dummies.fromfile(self.file_object,5)

                # NAMNF: The number of free atoms
                self.namnf = self._read_variable('i')

                # DELTA: The timestep
                delta = self._read_variable('f')

                # 10 blanc integers
                dummies = array.array('i')
                dummies.fromfile(self.file_object,10)

                # The end size of the first block
                input_integer = self._read_variable('i')
                if input_integer != 84 :
                        print('BAD DCD FORMAT')
                        raise PlamsError

                ############################################################

                # The size of the next block
                input_integer = self._read_variable('i')
        
                if (input_integer-4)%80 == 0 :
                        # NTITLE: The number of 80 char title strings
                        ntitle = self._read_variable('i')

                        for i in range(ntitle) :
                                car = self.file_object.read(80).decode()
                                #print car

                        # The end size of this block
                        input_integer = self._read_variable('i')
                else :
                        print('BAD DCD FORMAT')
                        raise PlamsError

                ############################################################

                # This should be an int32
                input_integer = self._read_variable('i')
                if input_integer != 4 :
                        print('BAD DCD FORMAT')
                        raise PlamsError

                # Read in the number of atoms
                self.ntap = self._read_variable('i')
                #print 'ntap: ',self.ntap
                if self.coords.shape == (0,3) :
                        self.coords = numpy.zeros((self.ntap,3))

                # This should be an int32
                input_integer = self._read_variable('i')
                if input_integer != 4 :
                        print('BAD DCD FORMAT')
                        raise PlamsError
                
                if self.namnf != 0 :
                        # This should be an int32
                        input_integer = self._read_variable('i')
                        if input_integer != (self.ntap-self.namnf)*4 :
                                print('BAD DCD FORMAT')
                                raise PlamsError
                
                        # This should be an int32
                        dummies = array.array('i')
                        dummies.fromfile(self.file_object,(self.ntap-self.namnf))
                        #print 'list',list(dummies)

                        # This should be an int32
                        input_integer = self._read_variable('i')
                        if input_integer != (self.ntap-self.namnf)*4 :
                                print('BAD DCD FORMAT')
                                raise PlamsError

                self.stepsize += 3 * self.ntap * self.floatsize
                self.stepsize += 5 * self.intsize

                if len(self.elements) != self.ntap :
                        self.elements = ['H']*self.ntap

        def read_next (self,molecule=None,read=True) :
                """
                Reads coordinates and lattice info from the current position of the cursor and returns it

                * ``molecule`` -- |Molecule| object in which the new data needs to be stored
                * ``read``     -- If set to False the cursor will move to the next frame without reading
                """
                # Check for end of file
                # This should be an int32
                input_integer_char = self.file_object.read(self.intsize)
                if len(input_integer_char) > 0 :
                        if self.firsttime :
                                self.celldata = False
                                input_integer = struct.unpack(self.byte_order+'i',input_integer_char)[0]
                                if input_integer == 48 :
                                        self.celldata = True
                else :
                        return None, None

                if not read and not self.firsttime :
                        return self._move_cursor_without_reading()
               
                # Read cell data 
                cell = numpy.zeros((3,3))
                if self.celldata :
                        cell_array = array.array('d')
                        try :
                                cell_array.fromfile(self.file_object,6)
                                if self.byte_order != '@' :
                                        cell_array.byteswap()
                        except EOFError :
                                return None, None
                        cell[numpy.tril_indices(3)] = cell_array
                        self.file_object.read(self.floatsize)

                        # Coords
                        if self.firsttime :
                                input_integer = self._read_variable('i')
                        else :
                                self.file_object.read(self.intsize)

                if self.firsttime :
                        if input_integer != 4*self.ntap :
                                print('BAD DCD FORMAT')
                                raise PlamsError
                
                xcoords = array.array('f')
                try :
                        xcoords.fromfile(self.file_object,self.ntap)
                        #xcoords = numpy.fromfile(self.file,dtype='f',count=self.ntap)
                        if self.byte_order != '@' :
                                xcoords.byteswap()
                except EOFError :
                        return None, None
              
                for i in range(2) : 
                        if self.firsttime :
                                input_integer = self._read_variable('i') 
                                if input_integer != 4*self.ntap :
                                        print('BAD DCD FORMAT')
                                        raise PlamsError
                        else :
                                self.file_object.read(self.intsize)

                ycoords = array.array('f')
                try:
                        ycoords.fromfile(self.file_object,self.ntap)
                        if self.byte_order != '@' :
                                ycoords.byteswap()
                except EOFError :
                        return None, None

                for i in range(2) :
                        if self.firsttime :
                                input_integer = self._read_variable('i')
                                if input_integer != 4*self.ntap :
                                        print('BAD DCD FORMAT')
                                        raise PlamsError
                        else :
                                self.file_object.read(self.intsize)

                zcoords = array.array('f')
                try :
                        zcoords.fromfile(self.file_object,self.ntap)
                        if self.byte_order != '@' :
                                zcoords.byteswap()
                except EOFError :
                        return None, None
              
                if self.firsttime : 
                        input_integer = self._read_variable('i') 
                        if input_integer != 4*self.ntap :
                                print('BAD DCD FORMAT')
                                raise PlamsError
                else :
                        self.file_object.read(self.intsize)

                # Check that self.coords has not been changed
                if not self.coords.shape == (self.ntap,3) :
                        raise PlamsError('The coordinate array has been changed outside the class')
                self.coords[:,0] = xcoords
                self.coords[:,1] = ycoords
                self.coords[:,2] = zcoords

                if self.firsttime :
                        if self.celldata :
                                self.stepsize += 6 * struct.calcsize('d')
                                self.stepsize += self.floatsize
                                self.stepsize += self.intsize
                        self.firsttime = False

                if isinstance(molecule,Molecule) :
                        self._set_plamsmol(self.coords, cell, molecule)

                self.position += 1
              
                return self.coords, cell

        def _is_endoffile (self) :
                """
                If the end of file is reached, return coords and cell as None
                """
                test = self.file_object.read(self.stepsize)
                return len(test) < self.stepsize

        def _write_header (self, ntap=0) :
                """
                Write the header of the DCD file
                """
                self.ntap = ntap
                self.coords = numpy.zeros((self.ntap,3))

                # This should be an int32
                self._write_variable(84,'i')

                # Read CORD string
                self.file_object.write('CORD'.encode())

                # Here we read the data of the timesteps
                # NSET: The number of sets of coordinates
                # ISTART: The starting timestep
                # NSAVC: The number of timesteps between dcd saves
                timesteps = array.array('i',[0,1,1])
                self.file_object.write(timesteps)

                # 5 blanc integers
                dummies = array.array('i',[0,0,0,0,0])
                self.file_object.write(dummies)

                # NAMNF: The number of free atoms
                self._write_variable(0,'i')

                # DELTA: The timestep
                delta = self._write_variable(self.delta,'f')

                # 10 blanc integers
                dummies = array.array('i')
                dummies.append(1)
                for i in range(8) :
                        dummies.append(0)
                dummies.append(24)
                self.file_object.write(dummies)

                # The end size of the first block
                self._write_variable(84,'i')

                ############################################################

                # The size of the next block
                self._write_variable(84,'i')
       
                ntitle = 1
                self._write_variable(ntitle,'i')
                title = 'REMARK Created with py_md'
                nlet = len(title)
                for i in range(80-nlet) :
                        title += ' '
                self.file_object.write(title.encode())

                self._write_variable(84,'i')
 
                ############################################################

                # This should be an int32
                self._write_variable(4,'i')

                # Read in the number of atoms
                self._write_variable(self.ntap,'i')

                # This should be an int32
                self._write_variable(4,'i')

        def write_next (self, coords=None, molecule=None, cell=[0.,0.,0.], conect=None) :
                """
                Write frame to next position in trajectory file

                * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates
                * ``molecule`` -- A molecule object to read the molecular data from
                * ``cell``     -- A set of lattice vectors or cell diameters
                * ``conect``   -- A dictionary containing connectivity info (not used)

                .. note::

                        Either ``coords`` or ``molecule`` are mandatory arguments
                """
                if isinstance(molecule,Molecule) :
                        coords, cell = self._read_plamsmol(molecule)[:2]
                cell = self._convert_cell(cell)

                # If this is the first step, write the header
                if self.position == 0 :
                        self._write_header(len(coords))

                if self.ntap != len(coords) :
                        print('BAD coordinate  list')
                        return

                # Invert the coords list [xcoords,ycoords,zcoords]
                #coords = numpy.array(coords).transpose()
                
                # This should be an int32
                self._write_variable(48,'i')
             
                cell = array.array('d',cell)
                self.file_object.write(cell)
                self._write_variable(0,'f')
                        
                # Coords
                self._write_variable(4*self.ntap,'i')
               
                xcoords = array.array('f',coords[:,0])
                self.file_object.write(xcoords)
                
                for i in range(2) : 
                        self._write_variable(4*self.ntap,'i')
               
                ycoords = array.array('f',coords[:,1])
                self.file_object.write(ycoords) 
                
                for i in range(2) :
                        self._write_variable(4*self.ntap,'i')
               
                zcoords = array.array('f',coords[:,2])
                self.file_object.write(zcoords) 

                self._write_variable(4*self.ntap,'i') 

                self.position += 1

        def _convert_cell (self, cell) :
                """
                Check format of cell, and convert to required lower triangle
                """
                if cell is None :
                        cell = numpy.zeros((3,3))[numpy.tril_indices(3)]
                elif len(cell)==3 and isinstance(cell[0],float) or isinstance(cell[0],numpy.float64) :
                        cell = [cell[0],0.,cell[1],0.,0.,cell[2]]
                elif len(cell) < 3 :
                        newcell = numpy.zeros((3,3))
                        newcell[:len(cell)] = cell
                        cell = newcell[numpy.tril_indices(3)]
                else :
                        cell = numpy.array(cell)
                        cell = cell[numpy.tril_indices(3)]
                return cell

        def _rewind_to_first_frame(self) :
                """
                Rewind the file to the first frame
                """
                self.file_object.seek(0)
                self.firsttime = True
                self.position = 0
                self.stepsize = 0
                self._read_header()

        def _rewind_n_frames(self,nframes) :
                """
                Rewind the file by nframes frames
                """
                whence = os.SEEK_CUR
                stepsize = self.stepsize + self.intsize
                self.file_object.seek(-nframes*stepsize,whence)
                self.position = self.position - nframes



# Add new results extraction method
@add_to_class(ADFFragmentResults)
def get_transfer_integrals(self):
    return self.job.full.results.read_rkf_section('TransferIntegrals', file='adf')

# Common settings for all 3 jobs
# In the interest of computational speed we use a minimal basis
# set. For more quantitatively meaningful results, you should use 
# a larger basis.
common = Settings()
common.input.ams.Task = 'SinglePoint'
common.input.adf.Basis.Type = 'SZ'
common.input.adf.Basis.Core = 'None'
common.input.adf.Symmetry = 'NoSym'

# Specific settings for full system job
full = Settings()
full.input.adf.transferintegrals = True

# Load XYZ file and separate it into 2 fragments
mol = Molecule('BenzeneDimer.xyz')
mol.guess_bonds()
fragments = mol.separate()
if len(fragments) != 2:
    log('ERROR: Molecule {} was split into {} fragments'.format(mol.properties.name, len(fragments)))
    import sys; sys.exit(1)
else:
    mol1, mol2 = fragments

# Alternatively one could simply load fragments from separate xyz files:
# mol1 = Molecule('fragment1.xyz')
# mol2 = Molecule('fragment2.xyz')

job = ADFFragmentJob(name='ADFTI',fragment1=mol1,fragment2=mol2,settings=common,full_settings=full)
results = job.run()

# TI is a dictionary with the whole TransferIntegrals section from adf.rkf
print('== Results ==')
TI = results.get_transfer_integrals()
for key, value in sorted(TI.items()):
    print('{:<28}: {:>12.6f}'.format(key, value))


import os
import inspect
import subprocess
from itertools import cycle

import numpy as np

try:
    import pandas as pd
    PANDAS = True
except ImportError:
    PANDAS = False

from .scmjob import (SCMJob, SCMResults)
from ...tools.units import Units

__all__ = ['CRSResults', 'CRSJob']


class CRSResults(SCMResults):
    """A |SCMResults| subclass for accessing results of |CRSJob|."""
    _kfext = '.crskf'
    _rename_map = {'CRSKF': '$JN.crskf'}

    @property
    def section(self) -> str:
        try:  # Return the cached value if possible
            return self._section
        except AttributeError:
            self._section = self.job.settings.input.property._h.upper()
            return self._section

    def get_energy(self, energy_type: str = "deltag", compound_idx: int = 0, unit: str = 'kcal/mol') -> float:
        """Returns the solute solvation energy from an Activity Coefficients calculation."""
        E = self.readkf(self.section, energy_type)[compound_idx]
        return Units.convert(E, 'kcal/mol', unit)

    def get_activity_coefficient(self, compound_idx: int = 0) -> float:
        """Return the solute activity coefficient from an Activity Coefficients calculation."""
        return self.readkf(self.section, 'gamma')[compound_idx]

    def get_sigma_profile(self, subsection: str = 'profil', as_df: bool = False) -> dict:
        r"""Grab all sigma profiles, returning a dictionary of Numpy Arrays.

        Values of :math:`\sigma` are stored under the ``"σ (e/A**2)"`` key.

        Results can be returned as a Pandas DataFrame by settings *as_df* to ``True``.

        The returned results can be plotted by passing them to the :meth:`CRSResults.plot` method.

        .. note::
            *as_df* = ``True`` requires the Pandas_ package.
            Plotting requires the `matplotlib <https://matplotlib.org/index.html>`__ package.

        .. _Pandas: https://pandas.pydata.org/

        """
        args = (subsection, 'σ (e/A**2)', 'chdval')
        try:
            return self._get_array_dict('SIGMAPROFILE', *args, as_df=as_df)
        except KeyError:
            return self._get_array_dict('PURESIGMAPROFILE', *args, as_df=as_df)

    def get_sigma_potential(self, subsection: str = 'mu', unit: str = 'kcal/mol',
                            as_df: bool = False) -> dict:
        r"""Grab all sigma profiles, expressed in *unit*, and return a dictionary of Numpy Arrays.

        Values of :math:`\sigma` are stored under the ``"σ (e/A**2)"`` key.

        Results can be returned as a Pandas DataFrame by settings *as_df* to ``True``.

        The returned results can be plotted by passing them to the :meth:`CRSResults.plot` method.

        .. note::
            *as_df* = ``True`` requires the Pandas_ package.
            Plotting requires the `matplotlib <https://matplotlib.org/index.html>`__ package.

        .. _Pandas: https://pandas.pydata.org/

        """
        args = (subsection, 'σ (e/A**2)', 'chdval')
        try:
            return self._get_array_dict('SIGMAPOTENTIAL', *args, unit=unit, as_df=as_df)
        except KeyError:
            return self._get_array_dict('PURESIGMAPOTENTIAL', *args, unit=unit, as_df=as_df)

    def get_prop_names(self, section = None) -> list:
        r"""Read the section of the .crskf file and return a list of the properties that were calculated.  The section argument can be supplied to look at previously-calculated results.  If no section name is supplied, the function defaults to using the most recent property that was calculated.

        """
        if section == None:
            section = self.section
        try:
            return self._kf.get_skeleton()[section]
        except KeyError:
            raise KeyError("Cannot find section name: " + str(section))

    def get_results(self, section = None) -> dict:
        r"""Read the section from the most recent calculation type and return the result as a dictionary.

        """
        if section == None:
            section = self.section

        output = getattr(self, '_prop_dict', False)
        if output and output["section"] == section:
            return output

        props = self.get_prop_names()
        try:
            props.remove("ncomp")
            props.remove("nitems")
        except ValueError:
            raise ValueError("Results object is missing or incomplete.")

        # first get the two ranges for the indices
        ncomp  = self.readkf(section, 'ncomp')
        nitems = self.readkf(section, 'nitems')

        np_dict = { "section" : section }
        for prop in props:
            tmp = self.readkf(section,prop)
            if prop == "filename":
                np_dict[prop] = [str(x).strip() for x in tmp.split()]
                continue
            if not isinstance(tmp,list):
                np_dict[prop] = tmp
            else:
                np_dict[prop] = np.array(tmp)
                if len(tmp) == ncomp*nitems:
                    np_dict[prop].shape = (ncomp,nitems)

        setattr(self, '_prop_dict', np_dict)
        return np_dict

    def plot(self, *arrays: np.ndarray, x_axis: str = None, plot_fig: bool = True, x_label = None, y_label = None):
        """Plot, show and return a series of COSMO-RS results as a matplotlib Figure instance.

        Accepts the output of, *e.g.*, :meth:`CRSResults.get_sigma_profile`:
        A dictionary of Numpy arrays or a Pandas DataFrame.

        Returns a matplotlib Figure_ instance which can be further modified to the users liking.
        Automatic plotting of the resulting figure can be disabled with the *plot_fig* argument.

        .. note::
            This method requires the `matplotlib <https://matplotlib.org/index.html>`__ package.

        .. note::
            The name of the dictionary/DataFrame key containing the index (*i.e.* the x-axis) can,
            and should, be manually specified in *x_axis* if a custom *x_axis* is passed
            to :meth:`CRSResults._get_array_dict`.
            This argument can be ignored otherwise.

        .. _Figure: https://matplotlib.org/api/_as_gen/matplotlib.figure.Figure.html#matplotlib.figure.Figure

        """  # noqa
        def get_x_axis(array, x_axis):
            """Find and return the index and its name."""
            if x_axis is None:
                return np.arange(array.shape[1])

            if isinstance(x_axis, str):
                ret = self._prop_dict[x_axis]
            else:
                ret = np.array(x_axis, copy=False)
            ret = ret.ravel()  # Flatten it
            return ret[:array.shape[1]]

        # Check if matplotlib is installed
        try:
            import matplotlib
            matplotlib.use('TkAgg') if plot_fig else matplotlib.use('Agg')
            import matplotlib.pyplot as plt
        except ImportError:
            method = self.__class__.__name__ + '.plot'
            raise ImportError("{}: this method requires the 'matplotlib' package".format(method))

        self.get_results()

        # Create a dictionary of 1d arrays
        array_dict = {}
        for array in arrays:
            name = None
            if isinstance(array, str):  # Array refers to a section in the kf file
                name = array
                array = self._prop_dict[array]

            # Ensure it's a 2D array
            array = np.array(array, ndmin=2, dtype=float, copy=False)

            # Fill the array dict with 1d arrays
            base_key = '' if name is None else name + ' '
            iterator = enumerate(array, 1) if array.shape[0] != 1 else zip(cycle(' '), array)
            for i, array_1d in iterator:
                key = f'{base_key}{i}'
                array_dict[key] = array_1d
        # Retrieve the index and its name
        index = get_x_axis(array, x_axis)
        # print ("INDEX::::", index)
        if x_label is None:
            if isinstance(x_axis,str):
                x_label = x_axis
            else:
                x_label = ""

        if y_label is None:
            y_label = ""

        # Assign various series to the plot
        fig, ax = plt.subplots()
        for k, v in array_dict.items():
            ax.plot(index, v, label=k)

        # Add the legend and x-label
        ax.legend()
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)

        # Show and return
        if plot_fig:
            plt.show()
        return fig

    def _get_array_dict(self, section: str, subsection: str, x_axis: str, index_subsection: str,
                        unit: str = 'kcal/mol', as_df: bool = False) -> dict:
        """Create dictionary or DataFrame containing all values in *section*/*subsection*.

        Takes the following arguments:
            * The *section*/*subsection* of the desired quantity.
            * The desired name of the index (*x_axis*).
            * The name of subsection containing the index (*index_subsection*).
            * The *unit* of the output quanty (ignore this keyword if not applicable).
            * If the result should be returned as Pandas DataFrame (*as_df*).

        """
        ret = self._construct_array_dict(section, subsection, unit)

        # Create the index
        index = self.readarray(section, index_subsection, dtype=float)
        if section in ('BINMIXCOEF', 'COMPOSITIONLINE', 'TERNARYMIX'):
            ncomponent = 3 if section == 'TERNARYMIX' else 2
            index.shape = ncomponent, len(index) // ncomponent
            iterator = np.nditer(index.astype(str), flags=['external_loop'], order='F')
            ret[x_axis] = np.array([' / '.join(i for i in item) for item in iterator])
        else:
            ret[x_axis] = index

        # Return a dictionary of arrays or a DataFrame
        if not as_df:
            return ret
        else:
            return self._dict_to_df(ret, section, x_axis)

    def _construct_array_dict(self, section: str, subsection: str, unit: str = 'kcal/mol') -> dict:
        """Construct dictionary containing all values in *section*/*subsection*."""
        # Use filenames as keys
        _filenames = self.readkf(section, 'filename').split()
        filenames = [_filenames] if not isinstance(_filenames, list) else _filenames

        # Grab the keys and the number of items per key
        keys = [os.path.basename(key) for key in filenames] + ['Total']
        nitems = self.readkf(section, 'nitems')

        # Use sigma profiles/potentials as values
        ratio = Units.conversion_ratio('kcal/mol', unit)
        values = ratio * self.readarray(section, subsection, dtype=float)
        values.shape = len(values) // nitems, nitems

        ret = dict(zip(keys, values))
        try:
            ret['Total'] = self.readarray(section, subsection + 'tot', dtype=float)
        except KeyError:
            pass
        return ret

    @staticmethod
    def _dict_to_df(array_dict: dict, section: str, x_axis: str):
        """Attempt to convert a dictionary into a DataFrame."""
        if not PANDAS:
            method = inspect.stack()[2][3]
            raise ImportError("{}: as_df=True requires the 'pandas' package".format(method))

        index = pd.Index(array_dict.pop(x_axis), name=x_axis)
        df = pd.DataFrame(array_dict, index=index)
        df.columns.name = section.lower()
        return df


class CRSJob(SCMJob):
    """A |SCMJob| subclass intended for running COSMO-RS jobs."""
    _command = 'crs'
    _result_type = CRSResults
    _subblock_end = 'end'
    
    def __init__(self, **kwargs) -> None:
        """Initialize a :class:`CRSJob` instance."""
        super().__init__(**kwargs)
        self.settings.ignore_molecule = True

    @staticmethod
    def cos_to_coskf(filename: str) -> str:
        """Convert a .cos file into a .coskf file with the :code:`$AMSBIN/cosmo2kf` command.

        Returns the filename of the new .coskf file.

        """
        filename_out = filename + 'kf'
        try:
            amsbin = os.environ['AMSBIN']
        except KeyError:
            raise EnvironmentError("cos_to_coskf: Failed to load 'cosmo2kf' from '$AMSBIN/'; "
                                   "the 'AMSBIN' environment variable has not been set")

        args = [os.path.join(amsbin, 'cosmo2kf'), filename, filename_out]
        subprocess.run(args)
        return filename_out



from .scmjob import SCMJob, SCMResults

__all__ = ['FCFJob', 'FCFResults']


class FCFResults(SCMResults):
    _kfext = '.t61'
    _rename_map = {'TAPE61':'$JN'+_kfext}

    def get_molecule(self, *args, **kwargs):
        raise PlamsError('FCFResults do not support get_molecule() method. You can get molecules from job1 or job2')


class FCFJob(SCMJob):
    """A class representing calculation of Franck-Condon factors using ``fcf`` program.

    Two new attributes are introduced: ``inputjob1`` and ``inputjob2``. They are used to supply KF files from previous runs to ``fcf`` program. The value can either be a string with a path to KF file or an instance of any type of |SCMJob| or |SCMResults| (in this case the path to corresponding KF file will be extracted automatically). If the value of ``inputjob1`` or ``inputjob2`` is ``None``, no automatic handling occurs and user needs to manually supply paths to input jobs using proper keywords placed in ``myjob.settings.input`` (``STATES`` or ``STATE1`` and ``STATE2``).


    The resulting ``TAPE61`` file is renamed to ``jobname.t61``.
    """
    _result_type = FCFResults
    _command = 'fcf'
    _top = ['states', 'state1', 'state2']

    def __init__(self, inputjob1=None, inputjob2=None, **kwargs):
        SCMJob.__init__(self,**kwargs)
        self.inputjob1 = inputjob1
        self.inputjob2 = inputjob2

    def _serialize_mol(self):
        self.settings.input.state1 = self.inputjob1
        self.settings.input.state2 = self.inputjob2

    def _remove_mol(self):
        if 'state1' in self.settings.input:
            del self.settings.input.state1
        if 'state2' in self.settings.input:
            del self.settings.input.state2



import numpy as np
try:
    from scipy.spatial.distance import cdist
    scipy_present = True
except ImportError:
    scipy_present = False

from .units import Units

__all__ = ['rotation_matrix', 'axis_rotation_matrix', 'distance_array', 'angle','dihedral','cell_shape','cellvectors_from_shape']

def rotation_matrix(vec1, vec2):
    """
    Calculate the rotation matrix rotating *vec1* to *vec2*. Vectors can be any containers with 3 numerical values. They don't need to be normalized. Returns 3x3 numpy array.
    """
    vec1, vec2 = np.array(vec1), np.array(vec2)
    a = vec1/np.linalg.norm(vec1)
    b = vec2/np.linalg.norm(vec2)

    # avoid division by zero in case of antiparallel vectors
    if abs(1+np.dot(a,b)) < 1E-8:
        return -np.eye(3)
    
    v1,v2,v3 = np.cross(a,b)
    M = np.array([[0, -v3, v2], [v3, 0, -v1], [-v2, v1, 0]])
    return (np.identity(3) + M + np.dot(M,M)/(1+np.dot(a,b)))


def axis_rotation_matrix(vector, angle, unit='radian'):
    """
    Calculate the rotation matrix rotating along the *vector* by *angle* expressed in *unit*.

    *vector* can be any container with 3 numerical values. They don't need to be normalized. A positive angle denotes counterclockwise rotation, when looking along *vector*. Returns 3x3 numpy array.
    """

    vector /= np.linalg.norm(vector)
    v0, v1, v2 = vector

    W = np.array([[0, -v2, v1],
                  [v2, 0, -v0],
                  [-v1, v0, 0]])

    angle = Units.convert(angle, unit, 'radian')
    a1 = np.sin(angle)
    a2 = 1.0 - np.cos(angle)

    return np.identity(3) + a1 * W + a2 * W@W


def distance_array(array1, array2):
    """Calculates distance between each pair of points in *array1* and *array2*. Returns 2D ``numpy`` array.

    Uses fast ``cdist`` function if ``scipy`` is present, otherwise falls back to slightly slower ``numpy`` loop. Arguments should be 2-dimensional ``numpy`` arrays with the same second dimension. If *array1* is A x N and *array2* is B x N, the returned array is A x B.
    """
    return cdist(array1, array2) if scipy_present else np.array([np.linalg.norm(i - array2, axis=1) for i in array1])


def angle(vec1, vec2, result_unit='radian'):
    """Calculate an angle between vectors *vec1* and *vec2*.

    *vec1* and *vec2* should be iterable containers of length 3 (for example: tuple, list, numpy array). Values stored in them are expressed in Angstrom. Returned value is expressed in *result_unit*.

    This method requires all atomic coordinates to be numerical values, :exc:`~exceptions.TypeError` is raised otherwise.
    """
    vec1 = np.array([*vec1], dtype=float)
    vec2 = np.array([*vec2], dtype=float)

    num = np.dot(vec1, vec2)
    den = np.sqrt(((vec1)**2).sum()) * np.sqrt(((vec2)**2).sum())
    return Units.convert(np.arccos(num/den), 'radian', result_unit)


def dihedral(p1, p2, p3, p4, unit='radian'):
    """Calculate the value of diherdal angle formed by points *p1*, *p2*, *p3* and *p4* in a 3D space. Arguments can be any containers with 3 numerical values, also instances of |Atom|. Returned value is always non-negative, measures the angle clockwise (looking along *p2-p3* vector) and is expressed in *unit*."""
    p1 = np.array([*p1], dtype=float)
    p2 = np.array([*p2], dtype=float)
    p3 = np.array([*p3], dtype=float)
    p4 = np.array([*p4], dtype=float)

    b0 = p1 - p2
    b1 = p3 - p2
    b2 = p4 - p3

    b1 /= np.linalg.norm(b1)
    v = b0 - np.dot(b0, b1)*b1
    w = b2 - np.dot(b2, b1)*b1

    x = np.dot(v, w)
    y = np.dot(np.cross(b1, v), w)
    ret = np.arctan2(y, x)
    ret = 2*np.pi+ret if ret < 0 else ret
    return Units.convert(ret, 'radian', unit)

def cell_shape (lattice) :
    """
    Converts lattice vectors to lengths and angles (in radians)
    Sets internal cell size data, based on set of cell vectors.

    *cellvectors* is list containing three cell vectors (a 3x3 matrix)
    """
    lattice = np.asarray(lattice)
    a,b,c = np.sqrt((lattice**2).sum(axis=1))

    if a == 0. and b == 0. and c == 0. :
            return

    alpha,beta,gamma = (90.,90.,90.)

    if c != 0 :
            alpha = angle (lattice[1],lattice[2])
            beta  = angle (lattice[0],lattice[2])
    if b != 0 :
            gamma = angle (lattice[0],lattice[1])

    return [a,b,c,alpha,beta,gamma]

def cell_lengths(lattice, unit='angstrom'):
    """Return the lengths of the lattice vector. Returns a list with the same length as the number of lattice vector."""

    if lattice is None or len(lattice) == 0:
        raise ValueError('Cannot calculate cell_lengths for nonperiodic system')
    lattice = np.asarray(lattice)
    ret = np.sqrt((lattice**2).sum(axis=1)) * Units.conversion_ratio('angstrom', unit)
    return ret.tolist()

def cell_angles(lattice, unit='degree'):
    """Return the angles between lattice vectors.

    unit : str
        output unit

    For 2D systems, returns a list [gamma]

    For 3D systems, returns a list [alpha, beta, gamma]
    """
    ndim = len(lattice)

    if ndim < 2:
        raise ValueError('Cannot calculate cell_angles for fewer than 2 lattice vectors. Tried with {} lattice vectors'.format(ndim))

    gamma = angle(lattice[0], lattice[1], result_unit=unit)

    if ndim == 2:
        return [gamma]

    if ndim >= 3:
        alpha = angle(lattice[1], lattice[2], result_unit=unit)
        beta = angle(lattice[0], lattice[2], result_unit=unit)
        return [alpha, beta, gamma]


def cellvectors_from_shape (box) :
    """
    Converts lengths and angles (in radians) of lattice vectors to the lattice vectors 
    """
    a = box[0]
    b = box[1]
    c = box[2]
    alpha, beta, gamma = 90., 90., 90
    if len(box) == 6 :
        alpha = box[3]#*np.pi/180.
        beta = box[4]#*np.pi/180.
        gamma = box[5]#*np.pi/180.

    va = [a,0.,0.]
    vb = [b*np.cos(gamma),b*np.sin(gamma),0.]

    cx = c*np.cos(beta)
    cy = (np.cos(alpha) - np.cos(beta)*np.cos(gamma)) * c / np.sin(gamma)
    volume = 1 - np.cos(alpha)**2 - np.cos(beta)**2 - np.cos(gamma)**2
    volume += 2 * np.cos(alpha) * np.cos(beta) * np.cos(gamma)
    volume = np.sqrt(volume)
    cz = c * volume / np.sin(gamma)
    vc = [cx,cy,cz]

    lattice = [va,vb,vc]

    return lattice



from .scmjob import SCMJob, SCMResults

__all__ = ['DensfJob', 'DensfResults']


class DensfResults(SCMResults):
    _kfext = '.t41'
    _rename_map = {'TAPE41':'$JN'+_kfext}

    def get_molecule(self, *args, **kwargs):
        raise PlamsError('DensfResults do not support get_molecule() method. You can get molecule from inputjob')


class DensfJob(SCMJob):
    """A class representing calculation of molecular properties on a grid using ``densf`` program.

    A new attribute ``inputjob`` is introduced to supply KF file from previously run job. The value can either be a string with a path to KF file or an instance of any type of |SCMJob| or |SCMResults| (in this case the path to corresponding KF file will be extracted automatically). If the value of ``inputjob`` is ``None``, no automatic handling occurs and user needs to manually supply path to input job using ``INPUTFILE`` keyword placed in ``myjob.settings.input``.

    The resulting ``TAPE41`` file is renamed to ``jobname.t41``.
    """
    _result_type = DensfResults
    _command = 'densf'
    _top = ['inputfile', 'units']

    def __init__(self, inputjob=None, **kwargs):
        SCMJob.__init__(self, **kwargs)
        self.inputjob = inputjob

    def _serialize_mol(self):
        self.settings.input.inputfile = self.inputjob

    def _remove_mol(self):
        if 'inputfile' in self.settings.input:
            del self.settings.input.inputfile

    def check(self):
        try:
            grep = self.results.grep_file('$JN.err', 'NORMAL TERMINATION')
        except:
            return False
        return len(grep) > 0